from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Query:
    """Base AST node for search queries."""

    def __and__(self, other: "Query") -> "And":
        return And(self, other)

    def __or__(self, other: "Query") -> "Or":
        return Or(self, other)

    def __invert__(self) -> "Not":
        return Not(self)


@dataclass(frozen=True, slots=True)
class Field(Query):
    """Field match: field=value."""

    field: str
    value: str


@dataclass(frozen=True, slots=True)
class And(Query):
    """Logical AND of two queries."""

    left: Query
    right: Query


@dataclass(frozen=True, slots=True)
class Or(Query):
    """Logical OR of two queries."""

    left: Query
    right: Query


@dataclass(frozen=True, slots=True)
class Not(Query):
    """Logical NOT of a query."""

    operand: Query


@dataclass(frozen=True, slots=True)
class YearRange(Query):
    """Year range filter."""

    start: int | None = None
    end: int | None = None


@dataclass(frozen=True, slots=True)
class CitationRange(Query):
    """Citation count filter."""

    min: int | None = None
    max: int | None = None


def title(value: str) -> Field:
    return Field("title", value)


def abstract(value: str) -> Field:
    return Field("abstract", value)


def author(value: str) -> Field:
    return Field("author", value)


def keyword(value: str) -> Field:
    return Field("keyword", value)


def doi(value: str) -> Field:
    return Field("doi", value)


def fulltext(value: str) -> Field:
    return Field("fulltext", value)


def year(start: int | None = None, end: int | None = None) -> YearRange:
    return YearRange(start, end)


def citations(min: int | None = None, max: int | None = None) -> CitationRange:
    """Filter by citation count range.

    Args:
        min: Minimum citation count (inclusive).
        max: Maximum citation count (inclusive).

    Examples:
        citations(100)
        citations(100, 500)
        citations(min=50)
        citations(max=200)
    """
    return CitationRange(min, max)


def has_fulltext(query: Query) -> bool:
    """Check if query contains a fulltext field.

    Args:
        query: The query AST to check.

    Returns:
        True if the query contains a Field with field="fulltext".
    """
    match query:
        case Field(field="fulltext"):
            return True
        case And(left=l, right=r) | Or(left=l, right=r):
            return has_fulltext(l) or has_fulltext(r)
        case Not(operand=o):
            return has_fulltext(o)
        case _:
            return False


def extract_fulltext_term(query: Query) -> str | None:
    """Extract the fulltext search term from query.

    Args:
        query: The query AST to extract from.

    Returns:
        The fulltext search term if found, None otherwise.
    """
    match query:
        case Field(field="fulltext", value=v):
            return v
        case And(left=l, right=r) | Or(left=l, right=r):
            return extract_fulltext_term(l) or extract_fulltext_term(r)
        case Not(operand=o):
            return extract_fulltext_term(o)
        case _:
            return None


def extract_citation_range(query: Query) -> CitationRange | None:
    """Extract CitationRange from query if present.

    Args:
        query: The query AST to check.

    Returns:
        CitationRange if found, None otherwise.
    """
    match query:
        case CitationRange() as cr:
            return cr
        case And(left=l, right=r) | Or(left=l, right=r):
            return extract_citation_range(l) or extract_citation_range(r)
        case Not(operand=o):
            return extract_citation_range(o)
        case _:
            return None


def remove_citation_range(query: Query) -> Query | None:
    """Remove CitationRange from query, returning the remaining query.

    Args:
        query: The query AST to transform.

    Returns:
        The query without CitationRange, or None if nothing remains.
    """
    match query:
        case CitationRange():
            return None
        case Field():
            return query
        case YearRange():
            return query
        case And(left=l, right=r):
            new_left = remove_citation_range(l)
            new_right = remove_citation_range(r)
            if new_left is None and new_right is None:
                return None
            if new_left is None:
                return new_right
            if new_right is None:
                return new_left
            return And(new_left, new_right)
        case Or(left=l, right=r):
            new_left = remove_citation_range(l)
            new_right = remove_citation_range(r)
            if new_left is None and new_right is None:
                return None
            if new_left is None:
                return new_right
            if new_right is None:
                return new_left
            return Or(new_left, new_right)
        case Not(operand=o):
            new_operand = remove_citation_range(o)
            if new_operand is None:
                return None
            return Not(new_operand)
        case _:
            return query


def remove_fulltext(query: Query) -> Query | None:
    """Remove fulltext field from query, returning the remaining query.

    Args:
        query: The query AST to transform.

    Returns:
        The query without fulltext fields, or None if nothing remains.
    """
    match query:
        case Field(field="fulltext"):
            return None
        case Field():
            return query
        case And(left=l, right=r):
            new_left = remove_fulltext(l)
            new_right = remove_fulltext(r)
            if new_left is None and new_right is None:
                return None
            if new_left is None:
                return new_right
            if new_right is None:
                return new_left
            return And(new_left, new_right)
        case Or(left=l, right=r):
            new_left = remove_fulltext(l)
            new_right = remove_fulltext(r)
            if new_left is None and new_right is None:
                return None
            if new_left is None:
                return new_right
            if new_right is None:
                return new_left
            return Or(new_left, new_right)
        case Not(operand=o):
            new_operand = remove_fulltext(o)
            if new_operand is None:
                return None
            return Not(new_operand)
        case _:
            return query



from __future__ import annotations

import pytest

from scimesh import search
from scimesh.providers import OpenAlex
from scimesh.query import author, title, year
from scimesh.query.combinators import citations
from tests.integration.conftest import (
    ATTENTION_PAPER_DOI,
    ATTENTION_PAPER_TITLE,
    ATTENTION_PAPER_YEAR,
    QueryExpectation,
    assert_paper_satisfies,
)


class TestOpenAlexSearchByTitle:
    @pytest.mark.asyncio
    async def test_title_filter_returns_matching_papers(self):
        q = title("transformer")
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(title_contains="transformer", source="openalex"),
            )

        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_title_filter_multiword(self):
        q = title("neural network")
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(title_contains="neural", source="openalex"),
            )

        assert len(papers) >= 1


class TestOpenAlexSearchByAuthor:
    @pytest.mark.asyncio
    async def test_author_filter_returns_matching_papers(self):
        q = author("Hinton")
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(author_contains="Hinton", source="openalex"),
            )

        assert len(papers) >= 1


class TestOpenAlexSearchByYear:
    @pytest.mark.asyncio
    async def test_year_range_filter(self):
        q = title("deep learning") & year(2020, 2022)
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(
                    title_contains="deep learning",
                    year_min=2020,
                    year_max=2022,
                    source="openalex",
                ),
            )

        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_year_min_only(self):
        q = title("machine learning") & year(2022, None)
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(year_min=2022, source="openalex"),
            )

        assert len(papers) >= 1


class TestOpenAlexSearchByCitations:
    @pytest.mark.asyncio
    async def test_citations_min_filter(self):
        q = title("deep learning") & citations(min=1000)
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(
                    citations_min=1000,
                    source="openalex",
                ),
            )

        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_citations_range_filter(self):
        q = title("neural network") & citations(min=100, max=500)
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(
                    citations_min=100,
                    citations_max=500,
                    source="openalex",
                ),
            )

        assert len(papers) >= 1


class TestOpenAlexSearchCombined:
    @pytest.mark.asyncio
    async def test_title_and_author(self):
        q = title("attention") & author("Vaswani")
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(
                    title_contains="attention",
                    author_contains="Vaswani",
                    source="openalex",
                ),
            )

        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_title_author_year_citations(self):
        q = title("learning") & year(2015, 2020) & citations(min=500)
        papers = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            papers.append(paper)
            assert_paper_satisfies(
                paper,
                QueryExpectation(
                    title_contains="learning",
                    year_min=2015,
                    year_max=2020,
                    citations_min=500,
                    source="openalex",
                ),
            )

        assert len(papers) >= 1


class TestOpenAlexGetByDoi:
    @pytest.mark.asyncio
    async def test_get_known_paper(self):
        openalex = OpenAlex()
        async with openalex:
            paper = await openalex.get(ATTENTION_PAPER_DOI)

        assert paper is not None
        assert ATTENTION_PAPER_TITLE.lower() in paper.title.lower()
        assert any("Vaswani" in a.name for a in paper.authors)
        assert paper.source == "openalex"

    @pytest.mark.asyncio
    async def test_get_returns_citations_count(self):
        openalex = OpenAlex()
        async with openalex:
            paper = await openalex.get(ATTENTION_PAPER_DOI)

        assert paper is not None
        assert paper.citations_count is not None
        assert paper.citations_count > 0


class TestOpenAlexCitations:
    @pytest.mark.asyncio
    async def test_citations_in(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1, "Attention paper should have citing papers"
        for paper in papers:
            assert paper.year >= ATTENTION_PAPER_YEAR, (
                f"Citing paper should be from {ATTENTION_PAPER_YEAR} or later, got {paper.year}"
            )

    @pytest.mark.asyncio
    async def test_citations_out(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="out",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1, "Attention paper should have references"
        for paper in papers:
            assert paper.year <= ATTENTION_PAPER_YEAR, (
                f"Reference should be from {ATTENTION_PAPER_YEAR} or earlier, got {paper.year}"
            )


class TestOpenAlexPagination:
    @pytest.mark.asyncio
    async def test_fetches_multiple_pages(self):
        q = title("machine learning")
        papers = []
        openalex = OpenAlex()

        async with openalex:
            count = 0
            async for paper in openalex.search(q):
                papers.append(paper)
                count += 1
                if count >= 250:
                    break

        assert len(papers) >= 200, "Should fetch more than one page"


class TestOpenAlexPaperFields:
    @pytest.mark.asyncio
    async def test_paper_has_required_fields(self):
        q = title("transformer")
        paper = None

        async for p in search(q, providers=[OpenAlex()], take=1):
            paper = p

        assert paper is not None
        assert paper.title
        assert paper.authors and len(paper.authors) > 0
        assert paper.year
        assert paper.source == "openalex"

    @pytest.mark.asyncio
    async def test_paper_has_doi(self):
        q = title("deep learning")
        papers_with_doi = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            if paper.doi:
                papers_with_doi.append(paper)

        assert len(papers_with_doi) > 0, "At least some papers should have DOI"

    @pytest.mark.asyncio
    async def test_paper_has_citations_count(self):
        q = title("neural network")
        papers_with_citations = []

        async for paper in search(q, providers=[OpenAlex()], take=10):
            if paper.citations_count is not None:
                papers_with_citations.append(paper)

        assert len(papers_with_citations) > 0, "Papers should have citations count"



from tests.integration.conftest import (
    ATTENTION_PAPER_DOI,
    ATTENTION_PAPER_TITLE,
    ATTENTION_PAPER_YEAR,
    BERT_PAPER_DOI,
    BERT_PAPER_TITLE,
    BERT_PAPER_YEAR,
    QueryExpectation,
    assert_paper_satisfies,
    assert_search_result_valid,
    assert_valid_paper,
    requires_scopus,
    requires_unpaywall,
)

__all__ = [
    "ATTENTION_PAPER_DOI",
    "ATTENTION_PAPER_TITLE",
    "ATTENTION_PAPER_YEAR",
    "BERT_PAPER_DOI",
    "BERT_PAPER_TITLE",
    "BERT_PAPER_YEAR",
    "QueryExpectation",
    "assert_paper_satisfies",
    "assert_search_result_valid",
    "assert_valid_paper",
    "requires_scopus",
    "requires_unpaywall",
]



"""Data models for workspace management using Pydantic."""

from datetime import datetime
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Discriminator, Field, Tag, TypeAdapter


class Constraints(BaseModel):
    """Search constraints for a workspace."""

    databases: list[str] = Field(default_factory=lambda: ["openalex", "semantic_scholar"])
    year_range: str | None = None


class Stats(BaseModel):
    """Statistics for a workspace."""

    total: int = 0
    with_pdf: int = 0
    included: int = 0
    excluded: int = 0
    maybe: int = 0
    unscreened: int = 0


class Framework(BaseModel):
    """Research question framework (PICO, SPIDER, or custom)."""

    type: str
    fields: dict[str, str] = Field(default_factory=dict)


class SLRWorkspace(BaseModel):
    """Systematic Literature Review workspace."""

    type: Literal["slr"] = "slr"
    question: str
    constraints: Constraints = Field(default_factory=Constraints)
    stats: Stats = Field(default_factory=Stats)
    framework: Framework
    inclusion: list[str] = Field(default_factory=list)
    exclusion: list[str] = Field(default_factory=list)


class ExplorationWorkspace(BaseModel):
    """Exploration workspace for broad literature surveys."""

    type: Literal["exploration"] = "exploration"
    question: str
    constraints: Constraints = Field(default_factory=Constraints)
    stats: Stats = Field(default_factory=Stats)
    limit: int | None = None
    started_at: datetime | None = None
    finished_at: datetime | None = None


class CollectionWorkspace(BaseModel):
    """Simple collection workspace for managing paper lists."""

    type: Literal["collection"] = "collection"
    question: str
    constraints: Constraints = Field(default_factory=Constraints)
    stats: Stats = Field(default_factory=Stats)


def _get_workspace_type(v: Any) -> str:
    """Extract workspace type for discriminated union."""
    if isinstance(v, dict):
        return v.get("type", "")
    return getattr(v, "type", "")


Workspace = Annotated[
    Annotated[SLRWorkspace, Tag("slr")]
    | Annotated[ExplorationWorkspace, Tag("exploration")]
    | Annotated[CollectionWorkspace, Tag("collection")],
    Discriminator(_get_workspace_type),
]

_workspace_adapter: TypeAdapter[Workspace] = TypeAdapter(Workspace)


def parse_workspace(
    data: dict[str, Any],
) -> SLRWorkspace | ExplorationWorkspace | CollectionWorkspace:
    """Parse a workspace dictionary into the appropriate workspace type.

    Args:
        data: Dictionary containing workspace data with a 'type' field.

    Returns:
        The appropriate workspace instance based on the type field.

    Raises:
        ValidationError: If the data is invalid or type is not recognized.
    """
    return _workspace_adapter.validate_python(data)


class SearchResults(BaseModel):
    """Results count for a search operation."""

    total: int
    unique: int


class Condensed(BaseModel):
    """Condensed paper information from exploration."""

    problem: str
    method: str
    results: str
    limitations: str | None = None
    relevance_to_exploration: str | None = None


class LogEntry(BaseModel):
    """Entry for a search/operation in the log."""

    id: str
    type: str
    query: str | None = None
    providers: list[str] = Field(default_factory=list)
    executed_at: datetime
    results: SearchResults | None = None
    seed_doi: str | None = None
    direction: str | None = None
    notes: str | None = None
    subtopics: list[str] = Field(default_factory=list)
    suggested_queries: list[str] = Field(default_factory=list)
    saturation: bool | None = None


class PaperEntry(BaseModel):
    """Entry for a paper in the papers index."""

    path: str
    doi: str
    title: str
    status: str = "unscreened"
    search_ids: list[str] = Field(default_factory=list)


class Screening(BaseModel):
    """Screening decision for a paper."""

    status: str
    reason: str
    screened_at: datetime


class PaperIndex(BaseModel):
    """Full index/metadata for a paper."""

    title: str
    authors: list[str]
    year: int
    doi: str | None = None
    sources: list[str] = Field(default_factory=list)
    urls: dict[str, str] = Field(default_factory=dict)
    tags: list[str] = Field(default_factory=list)
    citations: int | None = None
    journal: str | None = None
    open_access: bool | None = None
    pdf: str | None = None
    abstract: str | None = None
    screening: Screening | None = None
    subtopic: str | None = None
    relevance: str | None = None
    condensed: Condensed | None = None



# tests/test_download_base.py
import pytest

from scimesh.download import Downloader


def test_downloader_is_abstract():
    """Test that Downloader cannot be instantiated directly."""
    with pytest.raises(TypeError):
        Downloader()  # type: ignore[abstract]


def test_downloader_requires_download_method():
    """Test that subclass must implement download method."""

    class IncompleteDownloader(Downloader):
        name = "incomplete"

    with pytest.raises(TypeError):
        IncompleteDownloader()  # type: ignore[abstract]



"""YAML repository for workspace persistence."""

from datetime import UTC, datetime
from pathlib import Path

import yaml

from scimesh.workspace.models import (
    CollectionWorkspace,
    ExplorationWorkspace,
    LogEntry,
    PaperEntry,
    PaperIndex,
    SLRWorkspace,
    Stats,
    parse_workspace,
)


class WorkspaceNotFoundError(Exception):
    """Workspace not found."""


class WorkspaceExistsError(Exception):
    """Workspace already exists."""


class YamlWorkspaceRepository:
    """Repository for persisting workspace data to YAML files."""

    def __init__(self, root: Path):
        self.root = root

    @property
    def _index_path(self) -> Path:
        return self.root / "index.yaml"

    @property
    def _log_path(self) -> Path:
        return self.root / "log.yaml"

    @property
    def _papers_path(self) -> Path:
        return self.root / "papers.yaml"

    def load(self) -> SLRWorkspace | ExplorationWorkspace | CollectionWorkspace:
        if not self._index_path.exists():
            raise WorkspaceNotFoundError(f"Workspace not found at {self.root}")
        yaml_content = self._index_path.read_text()
        data = yaml.safe_load(yaml_content)
        return parse_workspace(data)

    def save(self, workspace: SLRWorkspace | ExplorationWorkspace | CollectionWorkspace) -> None:
        self.root.mkdir(parents=True, exist_ok=True)
        data = workspace.model_dump(mode="json", exclude_none=True)
        yaml_content = yaml.safe_dump(data, default_flow_style=False, allow_unicode=True)
        self._index_path.write_text(yaml_content)

    def exists(self) -> bool:
        return self._index_path.exists()

    def load_log(self) -> list[LogEntry]:
        if not self._log_path.exists():
            return []
        yaml_content = self._log_path.read_text()
        raw_list = yaml.safe_load(yaml_content) or []
        return [LogEntry.model_validate(entry) for entry in raw_list]

    def save_log(self, entries: list[LogEntry]) -> None:
        self.root.mkdir(parents=True, exist_ok=True)
        data = [entry.model_dump(mode="json", exclude_none=True) for entry in entries]
        yaml_content = yaml.safe_dump(data, default_flow_style=False, allow_unicode=True)
        self._log_path.write_text(yaml_content)

    def append_log(self, entry: LogEntry) -> None:
        entries = self.load_log()
        entries.append(entry)
        self.save_log(entries)

    def load_papers(self) -> list[PaperEntry]:
        if not self._papers_path.exists():
            return []
        yaml_content = self._papers_path.read_text()
        raw_list = yaml.safe_load(yaml_content) or []
        return [PaperEntry.model_validate(entry) for entry in raw_list]

    def save_papers(self, papers: list[PaperEntry]) -> None:
        self.root.mkdir(parents=True, exist_ok=True)
        data = [paper.model_dump(mode="json", exclude_none=True) for paper in papers]
        yaml_content = yaml.safe_dump(data, default_flow_style=False, allow_unicode=True)
        self._papers_path.write_text(yaml_content)

    def _paper_index_path(self, path: str) -> Path:
        return self.root / path / "index.yaml"

    def load_paper(self, path: str) -> PaperIndex:
        paper_path = self._paper_index_path(path)
        if not paper_path.exists():
            raise FileNotFoundError(f"Paper not found at {paper_path}")
        yaml_content = paper_path.read_text()
        data = yaml.safe_load(yaml_content)
        return PaperIndex.model_validate(data)

    def save_paper(self, path: str, paper: PaperIndex) -> None:
        paper_path = self._paper_index_path(path)
        paper_path.parent.mkdir(parents=True, exist_ok=True)
        data = paper.model_dump(mode="json", exclude_none=True)
        yaml_content = yaml.safe_dump(data, default_flow_style=False, allow_unicode=True)
        paper_path.write_text(yaml_content)

    def paper_exists(self, path: str) -> bool:
        return self._paper_index_path(path).exists()

    def set_paper_screening(self, paper_path: str, status: str, reason: str) -> None:
        """Set screening status for a paper.

        Args:
            paper_path: Relative path to paper (e.g., papers/2024/smith)
            status: included, excluded, or maybe
            reason: Reason for the decision
        """
        paper = self.load_paper(paper_path)
        now = datetime.now(UTC)

        paper_dict = paper.model_dump(mode="json", exclude_none=True)
        paper_dict["screening"] = {
            "status": status,
            "reason": reason,
            "screened_at": now.isoformat().replace("+00:00", "Z"),
        }

        paper_yaml_path = self.root / paper_path / "index.yaml"
        yaml_content = yaml.safe_dump(paper_dict, default_flow_style=False, allow_unicode=True)
        paper_yaml_path.write_text(yaml_content)

        papers = self.load_papers()
        updated = [
            PaperEntry(
                path=p.path,
                doi=p.doi,
                title=p.title,
                status=status if p.path == paper_path else p.status,
                search_ids=p.search_ids,
            )
            for p in papers
        ]
        self.save_papers(updated)

    def update_stats(self) -> Stats:
        """Recalculate and update workspace statistics."""
        workspace = self.load()
        papers = self.load_papers()

        total = 0
        with_pdf = 0
        included = 0
        excluded = 0
        maybe = 0
        unscreened = 0

        for entry in papers:
            paper_path = self.root / entry.path
            if not paper_path.exists():
                continue

            total += 1

            if (paper_path / "fulltext.pdf").exists():
                with_pdf += 1

            status = entry.status
            if status == "included":
                included += 1
            elif status == "excluded":
                excluded += 1
            elif status == "maybe":
                maybe += 1
            else:
                unscreened += 1

        new_stats = Stats(
            total=total,
            with_pdf=with_pdf,
            included=included,
            excluded=excluded,
            maybe=maybe,
            unscreened=unscreened,
        )

        workspace.stats = new_stats
        self.save(workspace)

        return new_stats



"""Workspace module for managing paper collections and SLR workflows."""

from scimesh.workspace.models import (
    CollectionWorkspace,
    Constraints,
    ExplorationWorkspace,
    Framework,
    LogEntry,
    PaperEntry,
    PaperIndex,
    Screening,
    SearchResults,
    SLRWorkspace,
    Stats,
    Workspace,
    parse_workspace,
)
from scimesh.workspace.repository import (
    WorkspaceExistsError,
    WorkspaceNotFoundError,
    YamlWorkspaceRepository,
)

__all__ = [
    "CollectionWorkspace",
    "Constraints",
    "ExplorationWorkspace",
    "Framework",
    "LogEntry",
    "PaperEntry",
    "PaperIndex",
    "Screening",
    "SearchResults",
    "SLRWorkspace",
    "Stats",
    "Workspace",
    "WorkspaceExistsError",
    "WorkspaceNotFoundError",
    "YamlWorkspaceRepository",
    "parse_workspace",
]



"""Fulltext search using SQLite FTS5."""

import sqlite3
from pathlib import Path


class FulltextIndex:
    """SQLite FTS5-based fulltext index for paper content.

    Stores extracted text from PDFs and allows fulltext search across
    indexed content.

    Attributes:
        db_path: Path to the SQLite database file.

    Example:
        >>> index = FulltextIndex()
        >>> index.add("10.1234/paper", "This is the paper content...")
        >>> results = index.search("paper content")
        >>> print(results)
        ['10.1234/paper']
    """

    def __init__(self, db_path: Path | None = None):
        """Initialize the fulltext index.

        Args:
            db_path: Path to the SQLite database file.
                Defaults to ~/.scimesh/fulltext.db
        """
        self.db_path = db_path or Path.home() / ".scimesh" / "fulltext.db"
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
        self._init_db()

    def _init_db(self) -> None:
        """Initialize the database schema."""
        with sqlite3.connect(self.db_path) as conn:
            conn.execute("""
                CREATE VIRTUAL TABLE IF NOT EXISTS papers_fts USING fts5(
                    paper_id,
                    content,
                    tokenize='porter unicode61'
                )
            """)

            conn.execute("""
                CREATE TABLE IF NOT EXISTS papers_meta (
                    paper_id TEXT PRIMARY KEY,
                    indexed_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    content_hash TEXT
                )
            """)
            conn.commit()

    def add(self, paper_id: str, content: str) -> None:
        """Add or update a paper's content in the index.

        Args:
            paper_id: The DOI or paper identifier.
            content: The extracted text content.
        """
        import hashlib

        content_hash = hashlib.sha256(content.encode()).hexdigest()

        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                "SELECT content_hash FROM papers_meta WHERE paper_id = ?",
                (paper_id,),
            )
            row = cursor.fetchone()

            if row and row[0] == content_hash:
                return

            if row:
                conn.execute(
                    "DELETE FROM papers_fts WHERE paper_id = ?",
                    (paper_id,),
                )
                conn.execute(
                    "UPDATE papers_meta SET content_hash = ?, indexed_at = CURRENT_TIMESTAMP "
                    "WHERE paper_id = ?",
                    (content_hash, paper_id),
                )
            else:
                conn.execute(
                    "INSERT INTO papers_meta (paper_id, content_hash) VALUES (?, ?)",
                    (paper_id, content_hash),
                )

            conn.execute(
                "INSERT INTO papers_fts (paper_id, content) VALUES (?, ?)",
                (paper_id, content),
            )
            conn.commit()

    def search(self, query: str, limit: int = 100) -> list[str]:
        """Search the index for matching papers.

        Args:
            query: The search query (supports FTS5 syntax).
            limit: Maximum number of results to return.

        Returns:
            List of matching paper IDs, ordered by relevance.
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                """
                SELECT paper_id, bm25(papers_fts) as rank
                FROM papers_fts
                WHERE papers_fts MATCH ?
                ORDER BY rank
                LIMIT ?
                """,
                (query, limit),
            )
            return [row[0] for row in cursor.fetchall()]

    def has(self, paper_id: str) -> bool:
        """Check if a paper is in the index.

        Args:
            paper_id: The DOI or paper identifier.

        Returns:
            True if the paper is indexed, False otherwise.
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                "SELECT 1 FROM papers_meta WHERE paper_id = ?",
                (paper_id,),
            )
            return cursor.fetchone() is not None

    def remove(self, paper_id: str) -> bool:
        """Remove a paper from the index.

        Args:
            paper_id: The DOI or paper identifier.

        Returns:
            True if the paper was removed, False if it wasn't in the index.
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                "DELETE FROM papers_fts WHERE paper_id = ?",
                (paper_id,),
            )
            conn.execute(
                "DELETE FROM papers_meta WHERE paper_id = ?",
                (paper_id,),
            )
            conn.commit()
            return cursor.rowcount > 0

    def count(self) -> int:
        """Get the number of indexed papers.

        Returns:
            Number of papers in the index.
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute("SELECT COUNT(*) FROM papers_meta")
            return cursor.fetchone()[0]

    def clear(self) -> None:
        """Clear all indexed content."""
        with sqlite3.connect(self.db_path) as conn:
            conn.execute("DELETE FROM papers_fts")
            conn.execute("DELETE FROM papers_meta")
            conn.commit()

    def list_papers(self, limit: int = 1000) -> list[str]:
        """List all indexed paper IDs.

        Args:
            limit: Maximum number of IDs to return.

        Returns:
            List of paper IDs.
        """
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                "SELECT paper_id FROM papers_meta ORDER BY indexed_at DESC LIMIT ?",
                (limit,),
            )
            return [row[0] for row in cursor.fetchall()]


def extract_text_from_pdf(pdf_path: Path) -> str | None:
    """Extract text from a PDF file.

    Attempts to use pdftotext (poppler) first, falls back to pypdf.

    Args:
        pdf_path: Path to the PDF file.

    Returns:
        Extracted text, or None if extraction failed.
    """
    import subprocess

    try:
        result = subprocess.run(
            ["pdftotext", "-enc", "UTF-8", str(pdf_path), "-"],
            capture_output=True,
            text=True,
            timeout=60,
        )
        if result.returncode == 0 and result.stdout.strip():
            return result.stdout
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass

    try:
        from pypdf import PdfReader  # pyright: ignore[reportMissingImports]

        reader = PdfReader(pdf_path)
        text_parts = []
        for page in reader.pages:
            text = page.extract_text()
            if text:
                text_parts.append(text)
        if text_parts:
            return "\n".join(text_parts)
    except ImportError:
        pass
    except Exception:
        pass

    return None


__all__ = ["FulltextIndex", "extract_text_from_pdf"]



from __future__ import annotations

import csv
import json
import re
import tempfile
from io import StringIO
from pathlib import Path

import pytest

from scimesh import collect_search, search
from scimesh.export import get_exporter
from scimesh.providers import Arxiv, OpenAlex
from scimesh.query import author, title, year


class TestExportJson:
    @pytest.mark.asyncio
    async def test_json_is_valid(self):
        q = title("transformer")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("json")
        output = exporter.to_string(result)

        data = json.loads(output)
        assert isinstance(data, list) or "papers" in data

    @pytest.mark.asyncio
    async def test_json_has_required_fields(self):
        q = title("neural network")
        result = await collect_search(search(q, providers=[OpenAlex()], take=10))

        exporter = get_exporter("json")
        output = exporter.to_string(result)
        data = json.loads(output)

        papers = data.get("papers", data) if isinstance(data, dict) else data
        assert len(papers) >= 1
        paper = papers[0]

        required_fields = ["title", "authors", "year"]
        for field in required_fields:
            assert field in paper, f"Missing field: {field}"

    @pytest.mark.asyncio
    async def test_json_preserves_query_data(self):
        q = title("self-attention") & author("Vaswani") & year(2017, 2020)
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("json")
        output = exporter.to_string(result)
        data = json.loads(output)

        papers = data.get("papers", data) if isinstance(data, dict) else data
        for paper_data in papers:
            assert "attention" in paper_data["title"].lower()
            author_names = " ".join(
                a["name"] if isinstance(a, dict) else str(a) for a in paper_data["authors"]
            ).lower()
            assert "vaswani" in author_names
            assert 2017 <= paper_data["year"] <= 2020

    @pytest.mark.asyncio
    async def test_json_to_file(self):
        q = title("deep learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "papers.json"
            exporter = get_exporter("json")
            exporter.export(result, filepath)

            assert filepath.exists()
            data = json.loads(filepath.read_text())
            assert len(data) >= 1


class TestExportCsv:
    @pytest.mark.asyncio
    async def test_csv_is_valid(self):
        q = title("transformer")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("csv")
        output = exporter.to_string(result)

        reader = csv.DictReader(StringIO(output))
        rows = list(reader)

        assert len(rows) >= 1

    @pytest.mark.asyncio
    async def test_csv_has_required_columns(self):
        q = title("neural network")
        result = await collect_search(search(q, providers=[OpenAlex()], take=10))

        exporter = get_exporter("csv")
        output = exporter.to_string(result)

        reader = csv.DictReader(StringIO(output))
        fieldnames = reader.fieldnames or []

        required_columns = ["title", "year", "authors"]
        for col in required_columns:
            assert col in fieldnames, f"Missing column: {col}"

    @pytest.mark.asyncio
    async def test_csv_data_matches_query(self):
        q = title("self-attention") & year(2017, 2020)
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("csv")
        output = exporter.to_string(result)

        reader = csv.DictReader(StringIO(output))
        for row in reader:
            assert "attent" in row["title"].lower()
            year_val = int(row["year"])
            assert 2017 <= year_val <= 2020

    @pytest.mark.asyncio
    async def test_csv_to_file(self):
        q = title("machine learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "papers.csv"
            exporter = get_exporter("csv")
            exporter.export(result, filepath)

            assert filepath.exists()

            with open(filepath) as f:
                reader = csv.DictReader(f)
                rows = list(reader)
                assert len(rows) >= 1


class TestExportBibtex:
    @pytest.mark.asyncio
    async def test_bibtex_has_entries(self):
        q = title("transformer")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("bibtex")
        output = exporter.to_string(result)

        entry_pattern = r"@\w+\{"
        entries = re.findall(entry_pattern, output)
        assert len(entries) >= 1

    @pytest.mark.asyncio
    async def test_bibtex_has_required_fields(self):
        q = title("neural network")
        result = await collect_search(search(q, providers=[OpenAlex()], take=10))

        exporter = get_exporter("bibtex")
        output = exporter.to_string(result)

        assert "title" in output.lower()
        assert "author" in output.lower()
        assert "year" in output.lower()

    @pytest.mark.asyncio
    async def test_bibtex_entry_structure(self):
        q = title("self-attention") & author("Vaswani")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("bibtex")
        output = exporter.to_string(result)

        entry_match = re.search(r"@(\w+)\{([^,]+),", output)
        assert entry_match is not None, "No BibTeX entry found"

        entry_type = entry_match.group(1)
        assert entry_type in ("article", "misc", "inproceedings", "book", "techreport")

    @pytest.mark.asyncio
    async def test_bibtex_contains_query_data(self):
        q = title("deep learning") & year(2020, 2022)
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("bibtex")
        output = exporter.to_string(result)

        assert "deep learning" in output.lower() or "Deep Learning" in output
        assert "2020" in output or "2021" in output or "2022" in output

    @pytest.mark.asyncio
    async def test_bibtex_to_file(self):
        q = title("machine learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "papers.bib"
            exporter = get_exporter("bibtex")
            exporter.export(result, filepath)

            assert filepath.exists()
            content = filepath.read_text()
            assert "@" in content


class TestExportRis:
    @pytest.mark.asyncio
    async def test_ris_has_records(self):
        q = title("transformer")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("ris")
        output = exporter.to_string(result)

        assert output.count("TY  - ") >= 1
        assert output.count("ER  - ") >= 1

    @pytest.mark.asyncio
    async def test_ris_records_are_balanced(self):
        q = title("neural network")
        result = await collect_search(search(q, providers=[OpenAlex()], take=10))

        exporter = get_exporter("ris")
        output = exporter.to_string(result)

        ty_count = output.count("TY  - ")
        er_count = output.count("ER  - ")

        assert ty_count == er_count, f"Unbalanced RIS: {ty_count} TY vs {er_count} ER"

    @pytest.mark.asyncio
    async def test_ris_has_required_tags(self):
        q = title("self-attention")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("ris")
        output = exporter.to_string(result)

        assert "TI  - " in output, "Missing title tag"
        assert "PY  - " in output or "Y1  - " in output, "Missing year tag"
        assert "AU  - " in output, "Missing author tag"

    @pytest.mark.asyncio
    async def test_ris_contains_query_data(self):
        q = title("deep learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("ris")
        output = exporter.to_string(result)

        title_match = re.search(r"TI  - (.+)", output)
        assert title_match is not None
        assert (
            "deep learning" in title_match.group(1).lower()
            or "learning" in title_match.group(1).lower()
        )

    @pytest.mark.asyncio
    async def test_ris_to_file(self):
        q = title("machine learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        with tempfile.TemporaryDirectory() as tmpdir:
            filepath = Path(tmpdir) / "papers.ris"
            exporter = get_exporter("ris")
            exporter.export(result, filepath)

            assert filepath.exists()
            content = filepath.read_text()
            assert "TY  - " in content


class TestExportTree:
    @pytest.mark.asyncio
    async def test_tree_has_papers(self):
        q = title("transformer")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("tree")
        output = exporter.to_string(result)

        assert len(output) > 0
        assert "├──" in output or "└──" in output

    @pytest.mark.asyncio
    async def test_tree_shows_year(self):
        q = title("self-attention") & year(2017, 2020)
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("tree")
        output = exporter.to_string(result)

        assert "Year:" in output

    @pytest.mark.asyncio
    async def test_tree_shows_authors(self):
        q = title("neural network")
        result = await collect_search(search(q, providers=[OpenAlex()], take=10))

        exporter = get_exporter("tree")
        output = exporter.to_string(result)

        assert "Authors:" in output

    @pytest.mark.asyncio
    async def test_tree_shows_url(self):
        q = title("deep learning")
        result = await collect_search(search(q, providers=[Arxiv()], take=10))

        exporter = get_exporter("tree")
        output = exporter.to_string(result)

        assert "URL:" in output or "http" in output



"""Sci-Hub downloader for scientific papers."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Any

import httpx

from scimesh.download.base import Downloader

if TYPE_CHECKING:
    from scimesh.download.host_concurrency import HostSemaphores

logger = logging.getLogger(__name__)


class SciHubDownloader(Downloader):
    """Downloader that uses Sci-Hub to fetch PDFs.

    Sci-Hub domains change frequently, so this downloader tries
    multiple known domains in order until one succeeds.

    Hosts accessed:
        - sci-hub.se, sci-hub.st, sci-hub.ru (HTML page + PDF download)
    """

    name = "scihub"

    domains = [
        "sci-hub.se",
        "sci-hub.st",
        "sci-hub.ru",
    ]

    def __init__(self, host_semaphores: HostSemaphores | None = None):
        """Initialize the Sci-Hub downloader.

        Args:
            host_semaphores: Shared per-host semaphores for concurrency control.
        """
        super().__init__(host_semaphores=host_semaphores)

    def _extract_pdf_url(self, html: str) -> str | None:
        """Extract PDF URL from Sci-Hub HTML response.

        Looks for <embed> or <iframe> tags containing .pdf URLs.

        Args:
            html: The HTML content to parse.

        Returns:
            The PDF URL if found, None otherwise.
        """

        patterns = [
            r'<object[^>]+data\s*=\s*["\']([^"\']*\.pdf[^"\']*)["\']',
            r'href\s*=\s*["\']([^"\']*download[^"\']*\.pdf)["\']',
            r'<embed[^>]+src\s*=\s*["\']([^"\']*\.pdf[^"\']*)["\']',
            r'<iframe[^>]+src\s*=\s*["\']([^"\']*\.pdf[^"\']*)["\']',
        ]

        for pattern in patterns:
            match = re.search(pattern, html, re.IGNORECASE)
            if match:
                url = match.group(1)

                if url.startswith("//"):
                    url = "https:" + url
                return url

        return None

    async def _get(self, url: str, **kwargs: Any) -> httpx.Response:
        """Make an HTTP GET request with per-host concurrency control.

        Args:
            url: The URL to fetch.
            **kwargs: Additional arguments for httpx.get().

        Returns:
            The HTTP response.
        """
        if self._client is None:
            raise RuntimeError("Downloader must be used as async context manager")

        if self._host_semaphores:
            async with self._host_semaphores.acquire(url):
                return await self._client.get(url, **kwargs)
        return await self._client.get(url, **kwargs)

    async def _attempt_download(self, doi: str, domain: str) -> bytes | None:
        """Attempt to download PDF from a specific Sci-Hub domain.

        Args:
            doi: The DOI of the paper.
            domain: The Sci-Hub domain to try.

        Returns:
            PDF bytes if successful, None otherwise.
        """
        if self._client is None:
            raise RuntimeError("Downloader must be used as async context manager")

        try:
            url = f"https://{domain}/{doi}"
            response = await self._get(url, follow_redirects=True)

            if response.status_code != 200:
                return None

            pdf_url = self._extract_pdf_url(response.text)
            if not pdf_url:
                preview = response.text[:500].replace("\n", " ")
                logger.debug("  [SciHub] No PDF URL in HTML from %s: %s...", domain, preview)
                return None

            if pdf_url.startswith("/"):
                pdf_url = f"https://{domain}{pdf_url}"

            pdf_response = await self._get(pdf_url, follow_redirects=True)
            if pdf_response.status_code == 200:
                return pdf_response.content

        except (httpx.RequestError, httpx.TimeoutException) as e:
            logger.debug("  [SciHub] Connection error for %s: %s", domain, e)
            return None

        return None

    async def download(self, doi: str) -> bytes | None:
        """Download PDF for the given DOI from Sci-Hub.

        Tries multiple Sci-Hub domains in order until one succeeds.

        Args:
            doi: The DOI of the paper to download.

        Returns:
            PDF bytes if found, None otherwise.
        """
        if self._client is None:
            raise RuntimeError("Downloader must be used as async context manager")

        logger.info("  [SciHub] Trying DOI: %s", doi)
        for domain in self.domains:
            result = await self._attempt_download(doi, domain)
            if result is not None:
                logger.info("  [SciHub] Success via %s", domain)
                return result
            logger.debug("  [SciHub] Failed via %s", domain)

        logger.info("  [SciHub] All domains failed for: %s", doi)
        return None



# tests/test_throttle.py
import asyncio
import time

import pytest

from scimesh.throttle import throttle


@pytest.mark.asyncio
async def test_throttle_limits_concurrent_calls():
    """Test that throttle limits concurrent calls to the specified number."""
    max_concurrent = 0
    current_concurrent = 0

    @throttle(calls=2, period=0.1)
    async def track_concurrent():
        nonlocal max_concurrent, current_concurrent
        current_concurrent += 1
        max_concurrent = max(max_concurrent, current_concurrent)
        await asyncio.sleep(0.05)  # Simulate some work
        current_concurrent -= 1
        return current_concurrent

    # Run 5 calls concurrently
    await asyncio.gather(*[track_concurrent() for _ in range(5)])

    # Should never exceed 2 concurrent calls
    assert max_concurrent <= 2


@pytest.mark.asyncio
async def test_throttle_enforces_minimum_interval():
    """Test that throttle enforces minimum time between calls."""
    call_times: list[float] = []

    @throttle(calls=1, period=0.1)
    async def track_time():
        call_times.append(time.monotonic())
        return len(call_times)

    # Make 3 sequential calls
    for _ in range(3):
        await track_time()

    # Check intervals between calls
    for i in range(1, len(call_times)):
        interval = call_times[i] - call_times[i - 1]
        # Allow small tolerance for timing
        assert interval >= 0.09, f"Interval {interval} is less than required 0.1s"


@pytest.mark.asyncio
async def test_multiple_decorated_functions_have_independent_state():
    """Test that each decorated function has its own throttle state."""
    func1_calls: list[float] = []
    func2_calls: list[float] = []

    @throttle(calls=1, period=0.2)
    async def slow_func():
        func1_calls.append(time.monotonic())
        return "slow"

    @throttle(calls=1, period=0.01)
    async def fast_func():
        func2_calls.append(time.monotonic())
        return "fast"

    # Call both functions concurrently multiple times
    start = time.monotonic()

    # Run slow_func twice and fast_func twice concurrently
    results = await asyncio.gather(
        slow_func(),
        fast_func(),
        slow_func(),
        fast_func(),
    )

    time.monotonic() - start

    assert results == ["slow", "fast", "slow", "fast"]

    # slow_func should have intervals of ~0.2s, fast_func should have ~0.01s
    # The functions should operate independently
    if len(func1_calls) >= 2:
        slow_interval = func1_calls[1] - func1_calls[0]
        assert slow_interval >= 0.19, f"slow_func interval {slow_interval} too short"

    if len(func2_calls) >= 2:
        fast_interval = func2_calls[1] - func2_calls[0]
        assert fast_interval < 0.1, f"fast_func interval {fast_interval} should be short"


@pytest.mark.asyncio
async def test_throttle_preserves_function_metadata():
    """Test that throttle preserves function name and docstring."""

    @throttle(calls=1, period=1.0)
    async def my_function():
        """This is my docstring."""
        return 42

    assert my_function.__name__ == "my_function"
    assert my_function.__doc__ == "This is my docstring."


@pytest.mark.asyncio
async def test_throttle_passes_arguments_correctly():
    """Test that throttle correctly passes args and kwargs to the function."""

    @throttle(calls=1, period=0.01)
    async def add(a, b, *, c=0):
        return a + b + c

    result = await add(1, 2, c=3)
    assert result == 6


@pytest.mark.asyncio
async def test_throttle_propagates_exceptions():
    """Test that exceptions from the decorated function are propagated."""

    @throttle(calls=1, period=0.01)
    async def raise_error():
        raise ValueError("Test error")

    with pytest.raises(ValueError, match="Test error"):
        await raise_error()



from abc import ABC, abstractmethod
from pathlib import Path

from scimesh.models import SearchResult


class Exporter(ABC):
    """Base class for result exporters."""

    @abstractmethod
    def to_string(self, result: SearchResult) -> str:
        """Export results to string."""
        ...

    def export(self, result: SearchResult, path: Path) -> None:
        """Export results to file."""
        path.write_text(self.to_string(result), encoding="utf-8")



# tests/test_scopus.py
from unittest.mock import MagicMock

import pytest

from scimesh.providers.scopus import Scopus
from scimesh.query.combinators import abstract, author, doi, fulltext, keyword, title, year


def test_translate_title():
    provider = Scopus()
    q = title("transformer")
    result = provider._translate_query(q)
    assert result == "TITLE(transformer)"


def test_translate_author():
    provider = Scopus()
    q = author("Vaswani")
    result = provider._translate_query(q)
    assert result == "AUTH(Vaswani)"


def test_translate_abstract():
    provider = Scopus()
    q = abstract("attention mechanism")
    result = provider._translate_query(q)
    assert result == "ABS(attention mechanism)"


def test_translate_keyword():
    provider = Scopus()
    q = keyword("machine learning")
    result = provider._translate_query(q)
    assert result == "KEY(machine learning)"


def test_translate_fulltext():
    provider = Scopus()
    q = fulltext("neural network")
    result = provider._translate_query(q)
    assert result == "ALL(neural network)"


def test_translate_doi():
    provider = Scopus()
    q = doi("10.1234/example")
    result = provider._translate_query(q)
    assert result == "DOI(10.1234/example)"


def test_translate_and():
    provider = Scopus()
    q = title("BERT") & author("Google")
    result = provider._translate_query(q)
    assert result == "(TITLE(BERT) AND AUTH(Google))"


def test_translate_or():
    provider = Scopus()
    q = title("BERT") | title("GPT")
    result = provider._translate_query(q)
    assert result == "(TITLE(BERT) OR TITLE(GPT))"


def test_translate_not():
    provider = Scopus()
    q = ~title("survey")
    result = provider._translate_query(q)
    assert result == "NOT TITLE(survey)"


def test_translate_year_range():
    provider = Scopus()
    q = year(2020, 2024)
    result = provider._translate_query(q)
    assert "PUBYEAR" in result
    # Should include years 2020-2024 inclusive
    assert "PUBYEAR > 2019" in result
    assert "PUBYEAR < 2025" in result


def test_translate_year_single():
    provider = Scopus()
    q = year(2023, 2023)
    result = provider._translate_query(q)
    assert result == "PUBYEAR = 2023"


def test_translate_year_start_only():
    provider = Scopus()
    q = year(start=2020)
    result = provider._translate_query(q)
    assert result == "PUBYEAR > 2019"


def test_translate_year_end_only():
    provider = Scopus()
    q = year(end=2024)
    result = provider._translate_query(q)
    assert result == "PUBYEAR < 2025"


def test_translate_complex_query():
    provider = Scopus()
    q = (title("deep learning") & author("LeCun")) | keyword("CNN")
    result = provider._translate_query(q)
    assert "TITLE(deep learning)" in result
    assert "AUTH(LeCun)" in result
    assert "KEY(CNN)" in result
    assert "AND" in result
    assert "OR" in result


def test_loads_api_key_from_env(monkeypatch):
    monkeypatch.setenv("SCOPUS_API_KEY", "test-key-123")
    provider = Scopus()
    assert provider._api_key == "test-key-123"


def test_explicit_api_key_overrides_env(monkeypatch):
    monkeypatch.setenv("SCOPUS_API_KEY", "env-key")
    provider = Scopus(api_key="explicit-key")
    assert provider._api_key == "explicit-key"


def test_no_api_key_returns_none():
    # Clear any existing env var
    import os

    original = os.environ.pop("SCOPUS_API_KEY", None)
    try:
        provider = Scopus()
        assert provider._api_key is None
    finally:
        if original:
            os.environ["SCOPUS_API_KEY"] = original


def test_provider_name():
    provider = Scopus()
    assert provider.name == "scopus"


def test_base_url():
    provider = Scopus()
    assert provider.BASE_URL == "https://api.elsevier.com/content/search/scopus"


def test_page_size_constant():
    """Scopus provider should have PAGE_SIZE constant of 25."""
    provider = Scopus()
    assert provider.PAGE_SIZE == 25


# Helper functions for mocking Scopus responses


def _make_scopus_entry(entry_id: int, title_text: str, year: int) -> dict:
    """Create a minimal Scopus entry object."""
    return {
        "dc:identifier": f"SCOPUS_ID:{entry_id}",
        "dc:title": title_text,
        "prism:coverDate": f"{year}-01-01",
        "dc:creator": "Test Author",
        "prism:doi": f"10.1234/test{entry_id}",
        "prism:publicationName": "Test Journal",
        "citedby-count": "10",
        "link": [
            {
                "@ref": "self",
                "@href": f"https://api.elsevier.com/content/abstract/scopus_id/{entry_id}",
            },
            {
                "@ref": "scopus",
                "@href": f"https://www.scopus.com/record/display.uri?eid={entry_id}",
            },
        ],
        "openaccessFlag": False,
    }


def _make_scopus_response(
    entries: list[dict],
    total: int,
    next_cursor: str | None,
    self_cursor: str = "*",
) -> dict:
    """Create a mock Scopus API response with pagination links."""
    links = [
        {
            "@ref": "self",
            "@href": f"https://api.elsevier.com/content/search/scopus?cursor={self_cursor}",
        },
    ]
    if next_cursor:
        links.append(
            {
                "@ref": "next",
                "@href": f"https://api.elsevier.com/content/search/scopus?cursor={next_cursor}",
            }
        )
    return {
        "search-results": {
            "opensearch:totalResults": str(total),
            "opensearch:startIndex": "0",
            "opensearch:itemsPerPage": str(len(entries)),
            "link": links,
            "entry": entries,
        }
    }


@pytest.mark.asyncio
async def test_search_paginates_with_cursor():
    """Scopus search should paginate using cursor-based pagination."""
    provider = Scopus(api_key="test-api-key")

    # Create 3 pages of results (25 + 25 + 10 = 60 total)
    page1_entries = [_make_scopus_entry(i, f"Paper {i}", 2023) for i in range(25)]
    page2_entries = [_make_scopus_entry(i, f"Paper {i}", 2023) for i in range(25, 50)]
    page3_entries = [_make_scopus_entry(i, f"Paper {i}", 2023) for i in range(50, 60)]

    page1_response = _make_scopus_response(page1_entries, 60, "cursor_page2", "*")
    page2_response = _make_scopus_response(page2_entries, 60, "cursor_page3", "cursor_page2")
    page3_response = _make_scopus_response(page3_entries, 60, None, "cursor_page3")

    call_count = 0
    requested_urls = []

    async def mock_get(url: str, headers=None):
        nonlocal call_count
        call_count += 1
        requested_urls.append(url)
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()

        if call_count == 1:
            mock_response.json = MagicMock(return_value=page1_response)
        elif call_count == 2:
            mock_response.json = MagicMock(return_value=page2_response)
        else:
            mock_response.json = MagicMock(return_value=page3_response)

        return mock_response

    mock_client = MagicMock()
    mock_client.get = mock_get

    provider._client = mock_client

    papers = []
    async for paper in provider.search(title("test")):
        papers.append(paper)

    # Should have fetched all 60 papers across 3 pages
    assert len(papers) == 60
    assert call_count == 3

    # Verify first request uses initial cursor "*"
    assert "cursor=%2A" in requested_urls[0] or "cursor=*" in requested_urls[0]
    # Verify subsequent requests use the cursor from previous response
    assert "cursor=cursor_page2" in requested_urls[1]
    assert "cursor=cursor_page3" in requested_urls[2]


@pytest.mark.asyncio
async def test_search_single_page_when_results_fit():
    """Scopus search should not paginate when results fit in single page."""
    provider = Scopus(api_key="test-api-key")

    entries = [_make_scopus_entry(i, f"Paper {i}", 2023) for i in range(10)]
    response = _make_scopus_response(entries, 10, None)

    call_count = 0

    async def mock_get(url: str, headers=None):
        nonlocal call_count
        call_count += 1
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()
        mock_response.json = MagicMock(return_value=response)
        return mock_response

    mock_client = MagicMock()
    mock_client.get = mock_get

    provider._client = mock_client

    papers = []
    async for paper in provider.search(title("test")):
        papers.append(paper)

    assert len(papers) == 10
    assert call_count == 1



from __future__ import annotations

import pytest

from scimesh.providers import OpenAlex, Scopus, SemanticScholar
from tests.integration.conftest import (
    ATTENTION_PAPER_DOI,
    ATTENTION_PAPER_YEAR,
    BERT_PAPER_DOI,
    BERT_PAPER_YEAR,
    assert_valid_paper,
    requires_scopus,
)


class TestOpenAlexCitationsIn:
    @pytest.mark.asyncio
    async def test_citations_in_returns_papers(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1, "Attention paper is highly cited"

    @pytest.mark.asyncio
    async def test_citations_in_year_constraint(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=20,
            ):
                papers.append(paper)

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year >= ATTENTION_PAPER_YEAR, (
                f"Citing paper must be from {ATTENTION_PAPER_YEAR} or later, got {paper.year}"
            )

    @pytest.mark.asyncio
    async def test_citations_in_has_valid_papers(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=5,
            ):
                papers.append(paper)

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.title
            assert paper.authors


class TestOpenAlexCitationsOut:
    @pytest.mark.asyncio
    async def test_citations_out_returns_papers(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="out",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1, "Attention paper has references"

    @pytest.mark.asyncio
    async def test_citations_out_year_constraint(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="out",
                max_results=20,
            ):
                papers.append(paper)

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year <= ATTENTION_PAPER_YEAR, (
                f"Reference must be from {ATTENTION_PAPER_YEAR} or earlier, got {paper.year}"
            )


class TestOpenAlexCitationsBoth:
    @pytest.mark.asyncio
    async def test_citations_both_returns_papers(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                ATTENTION_PAPER_DOI,
                direction="both",
                max_results=20,
            ):
                papers.append(paper)

        assert len(papers) >= 1

        years = [p.year for p in papers]
        has_before = any(y <= ATTENTION_PAPER_YEAR for y in years)
        has_after = any(y >= ATTENTION_PAPER_YEAR for y in years)

        assert has_before or has_after, "Should have papers from before or after"


class TestSemanticScholarCitations:
    @pytest.mark.asyncio
    async def test_citations_in_returns_papers(self):
        ss = SemanticScholar()
        papers = []

        async with ss:
            async for paper in ss.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        if len(papers) == 0:
            pytest.skip("SemanticScholar API returned no results (likely rate limited)")
        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_citations_in_year_constraint(self):
        ss = SemanticScholar()
        papers = []

        async with ss:
            async for paper in ss.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        if len(papers) == 0:
            pytest.skip("SemanticScholar API returned no results (likely rate limited)")

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year >= ATTENTION_PAPER_YEAR

    @pytest.mark.asyncio
    async def test_citations_out_returns_papers(self):
        ss = SemanticScholar()
        papers = []

        async with ss:
            async for paper in ss.citations(
                ATTENTION_PAPER_DOI,
                direction="out",
                max_results=10,
            ):
                papers.append(paper)

        if len(papers) == 0:
            pytest.skip("SemanticScholar API returned no results (likely rate limited)")
        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_citations_out_year_constraint(self):
        ss = SemanticScholar()
        papers = []

        async with ss:
            async for paper in ss.citations(
                ATTENTION_PAPER_DOI,
                direction="out",
                max_results=10,
            ):
                papers.append(paper)

        if len(papers) == 0:
            pytest.skip("SemanticScholar API returned no results (likely rate limited)")

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year <= ATTENTION_PAPER_YEAR


@requires_scopus
@pytest.mark.skip(reason="Requires Citation Overview API permissions")
class TestScopusCitations:
    @pytest.mark.asyncio
    async def test_citations_in_returns_papers(self):
        scopus = Scopus()
        papers = []

        async with scopus:
            async for paper in scopus.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1

    @pytest.mark.asyncio
    async def test_citations_in_year_constraint(self):
        scopus = Scopus()
        papers = []

        async with scopus:
            async for paper in scopus.citations(
                ATTENTION_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year >= ATTENTION_PAPER_YEAR


class TestCitationsForDifferentPapers:
    @pytest.mark.asyncio
    async def test_bert_citations_in(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                BERT_PAPER_DOI,
                direction="in",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year >= BERT_PAPER_YEAR

    @pytest.mark.asyncio
    async def test_bert_citations_out(self):
        openalex = OpenAlex()
        papers = []

        async with openalex:
            async for paper in openalex.citations(
                BERT_PAPER_DOI,
                direction="out",
                max_results=10,
            ):
                papers.append(paper)

        assert len(papers) >= 1

        for paper in papers:
            assert_valid_paper(paper)
            assert paper.year <= BERT_PAPER_YEAR



import logging
import re
from collections.abc import AsyncIterator, Iterable
from dataclasses import dataclass
from pathlib import Path

import httpx
import streamish as st

from scimesh.cache import PaperCache
from scimesh.download.base import Downloader
from scimesh.download.fallback import FallbackDownloader
from scimesh.download.host_concurrency import HostSemaphores
from scimesh.download.openaccess import OpenAccessDownloader
from scimesh.download.scihub import SciHubDownloader

try:
    from scimesh.download.playwright import PlaywrightConfig, PlaywrightDownloader

    PLAYWRIGHT_AVAILABLE = True
except ImportError:
    PlaywrightConfig = None  # type: ignore[misc,assignment]
    PlaywrightDownloader = None  # type: ignore[misc,assignment]
    PLAYWRIGHT_AVAILABLE = False

logger = logging.getLogger(__name__)


@dataclass
class DownloadResult:
    """Result of a paper download attempt."""

    doi: str
    success: bool
    filename: str | None = None
    source: str | None = None
    error: str | None = None


def make_filename(doi: str) -> str:
    """Sanitize DOI to create a valid filename.

    Replaces invalid filesystem characters with safe alternatives.

    Args:
        doi: The DOI to sanitize.

    Returns:
        A sanitized filename (without .pdf extension).

    Example:
        >>> make_filename("10.1234/paper.v1")
        '10.1234_paper.v1'
    """
    filename = doi.replace("/", "_")

    invalid_chars = r'[\\:*?"<>|]'
    filename = re.sub(invalid_chars, "_", filename)

    return filename


async def download_papers(
    dois: Iterable[str],
    output_dir: Path,
    downloaders: list[Downloader] | None = None,
    cache: PaperCache | None = None,
    use_cache: bool = True,
    max_concurrency: int = 5,
) -> AsyncIterator[DownloadResult]:
    """Download papers for a list of DOIs.

    Tries each downloader in order until one succeeds. Saves PDFs to the
    output directory with sanitized filenames based on DOIs.

    Args:
        dois: An iterable of DOIs to download.
        output_dir: Directory to save downloaded PDFs.
        downloaders: Optional list of Downloader instances to use.
        cache: Optional PaperCache instance.
        use_cache: Whether to use caching. Defaults to True.
        max_concurrency: Maximum number of concurrent downloads.

    Yields:
        DownloadResult for each DOI.
    """
    if downloaders is None:
        downloaders = [OpenAccessDownloader()]

    if use_cache and cache is None:
        cache = PaperCache()

    output_dir.mkdir(parents=True, exist_ok=True)

    async def download_one(doi: str) -> DownloadResult:
        return await _download_single(doi, output_dir, downloaders, cache if use_cache else None)

    async for result in st.map_async(download_one, dois, concurrency=max_concurrency):
        yield result


async def _download_single(
    doi: str,
    output_dir: Path,
    downloaders: list[Downloader],
    cache: PaperCache | None = None,
) -> DownloadResult:
    """Download a single paper, trying each downloader in order.

    If a cache is provided, checks the cache first and saves successful
    downloads to the cache.

    Args:
        doi: The DOI to download.
        output_dir: Directory to save the PDF.
        downloaders: List of downloaders to try.
        cache: Optional PaperCache for caching PDFs.

    Returns:
        DownloadResult indicating success or failure.
    """
    filename = make_filename(doi) + ".pdf"
    filepath = output_dir / filename

    if cache is not None:
        cached_path = cache.get_pdf_path(doi)
        if cached_path is not None:
            pdf_bytes = cached_path.read_bytes()
            filepath.write_bytes(pdf_bytes)
            return DownloadResult(
                doi=doi,
                success=True,
                filename=filename,
                source="cache",
            )

    for downloader in downloaders:
        try:
            async with downloader:
                pdf_bytes = await downloader.download(doi)
                if pdf_bytes is not None:
                    filepath.write_bytes(pdf_bytes)
                    if cache is not None:
                        cache.save_pdf(doi, pdf_bytes)
                    return DownloadResult(
                        doi=doi,
                        success=True,
                        filename=filename,
                        source=downloader.name,
                    )
        except httpx.TimeoutException:
            logger.debug("Timeout from %s for DOI %s", downloader.name, doi)
            continue
        except httpx.HTTPStatusError as e:
            logger.debug("HTTP %s from %s for DOI %s", e.response.status_code, downloader.name, doi)
            continue
        except httpx.RequestError as e:
            logger.debug("Request error from %s for DOI %s: %s", downloader.name, doi, e)
            continue
        except Exception as e:
            logger.warning("Unexpected error from %s for DOI %s: %s", downloader.name, doi, e)
            continue

    return DownloadResult(
        doi=doi,
        success=False,
        error=f"All downloaders failed for DOI: {doi}",
    )


__all__ = [
    "Downloader",
    "DownloadResult",
    "FallbackDownloader",
    "HostSemaphores",
    "OpenAccessDownloader",
    "PaperCache",
    "PLAYWRIGHT_AVAILABLE",
    "PlaywrightConfig",
    "PlaywrightDownloader",
    "SciHubDownloader",
    "download_papers",
    "make_filename",
]



# tests/test_cache.py
"""Tests for the PaperCache class."""

import pytest

from scimesh.cache import PaperCache
from scimesh.download import Downloader, download_papers


class TestPaperCache:
    """Tests for PaperCache initialization and directory creation."""

    def test_default_cache_dir(self, tmp_path, monkeypatch):
        """Test that default cache directory is in home directory."""
        # Mock Path.home() to return tmp_path
        monkeypatch.setattr("pathlib.Path.home", lambda: tmp_path)

        cache = PaperCache()

        assert cache.cache_dir == tmp_path / ".scimesh" / "cache"
        assert cache.pdf_dir == tmp_path / ".scimesh" / "cache" / "pdfs"
        assert cache.text_dir == tmp_path / ".scimesh" / "cache" / "text"

    def test_custom_cache_dir(self, tmp_path):
        """Test using custom cache directory."""
        custom_dir = tmp_path / "my_cache"
        cache = PaperCache(cache_dir=custom_dir)

        assert cache.cache_dir == custom_dir
        assert cache.pdf_dir == custom_dir / "pdfs"
        assert cache.text_dir == custom_dir / "text"

    def test_directories_created_on_init(self, tmp_path):
        """Test that cache directories are created on initialization."""
        cache_dir = tmp_path / "new_cache"
        assert not cache_dir.exists()

        cache = PaperCache(cache_dir=cache_dir)

        assert cache.pdf_dir.exists()
        assert cache.text_dir.exists()


class TestMakeSafeFilename:
    """Tests for _make_safe_filename method."""

    def test_basic_doi(self, tmp_path):
        """Test basic DOI sanitization."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache._make_safe_filename("10.1234/paper.v1") == "10.1234_paper.v1"

    def test_slash_replacement(self, tmp_path):
        """Test that slashes are replaced with underscores."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache._make_safe_filename("10.1234/abc/def") == "10.1234_abc_def"

    def test_backslash_replacement(self, tmp_path):
        """Test that backslashes are replaced."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache._make_safe_filename("10.1234\\paper") == "10.1234_paper"

    def test_colon_replacement(self, tmp_path):
        """Test that colons are replaced."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache._make_safe_filename("10.1234:paper") == "10.1234_paper"

    def test_special_characters_replacement(self, tmp_path):
        """Test that special characters are replaced."""
        cache = PaperCache(cache_dir=tmp_path)
        # Asterisk
        assert cache._make_safe_filename("10.1234*paper") == "10.1234_paper"
        # Question mark
        assert cache._make_safe_filename("10.1234?paper") == "10.1234_paper"
        # Double quotes
        assert cache._make_safe_filename('10.1234"paper') == "10.1234_paper"
        # Angle brackets
        assert cache._make_safe_filename("10.1234<paper>") == "10.1234_paper_"
        # Pipe
        assert cache._make_safe_filename("10.1234|paper") == "10.1234_paper"

    def test_multiple_invalid_chars(self, tmp_path):
        """Test DOI with multiple invalid characters."""
        cache = PaperCache(cache_dir=tmp_path)
        result = cache._make_safe_filename('10.1234/pa:per*v1?"test"')
        assert result == "10.1234_pa_per_v1__test_"

    def test_real_world_dois(self, tmp_path):
        """Test with real-world DOI formats."""
        cache = PaperCache(cache_dir=tmp_path)
        # Nature article DOI format
        assert cache._make_safe_filename("10.1038/nature12373") == "10.1038_nature12373"
        # PLoS ONE DOI format
        assert (
            cache._make_safe_filename("10.1371/journal.pone.0123456")
            == "10.1371_journal.pone.0123456"
        )
        # arXiv DOI format
        assert cache._make_safe_filename("10.48550/arXiv.2301.00001") == "10.48550_arXiv.2301.00001"

    def test_very_long_doi_uses_hash(self, tmp_path):
        """Test that very long DOIs are handled with hashing."""
        cache = PaperCache(cache_dir=tmp_path)
        long_doi = "10.1234/" + "a" * 300
        result = cache._make_safe_filename(long_doi)
        # Should be truncated and have a hash suffix
        assert len(result) <= 200
        assert "_" in result[-20:]  # Hash suffix separated by underscore


class TestPDFCache:
    """Tests for PDF caching functionality."""

    def test_has_pdf_false_when_not_cached(self, tmp_path):
        """Test has_pdf returns False for uncached PDF."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache.has_pdf("10.1234/paper") is False

    def test_has_pdf_true_after_save(self, tmp_path):
        """Test has_pdf returns True after saving PDF."""
        cache = PaperCache(cache_dir=tmp_path)
        cache.save_pdf("10.1234/paper", b"%PDF-1.4 test content")
        assert cache.has_pdf("10.1234/paper") is True

    def test_get_pdf_path_returns_none_when_not_cached(self, tmp_path):
        """Test get_pdf_path returns None for uncached PDF."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache.get_pdf_path("10.1234/paper") is None

    def test_get_pdf_path_returns_path_after_save(self, tmp_path):
        """Test get_pdf_path returns correct path after saving."""
        cache = PaperCache(cache_dir=tmp_path)
        saved_path = cache.save_pdf("10.1234/paper", b"%PDF-1.4 test")

        retrieved_path = cache.get_pdf_path("10.1234/paper")
        assert retrieved_path is not None
        assert retrieved_path == saved_path
        assert retrieved_path.exists()

    def test_save_and_retrieve_pdf(self, tmp_path):
        """Test saving and retrieving PDF content."""
        cache = PaperCache(cache_dir=tmp_path)
        pdf_content = b"%PDF-1.4 test content with binary data \x00\x01\x02"

        saved_path = cache.save_pdf("10.1234/paper", pdf_content)

        # Verify file was saved correctly
        assert saved_path.exists()
        assert saved_path.read_bytes() == pdf_content
        assert saved_path.suffix == ".pdf"

    def test_save_pdf_returns_correct_path(self, tmp_path):
        """Test that save_pdf returns the correct path."""
        cache = PaperCache(cache_dir=tmp_path)
        path = cache.save_pdf("10.1234/paper", b"test")

        expected_path = cache.pdf_dir / "10.1234_paper.pdf"
        assert path == expected_path

    def test_save_pdf_overwrites_existing(self, tmp_path):
        """Test that save_pdf overwrites existing cached PDF."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_pdf("10.1234/paper", b"original content")
        cache.save_pdf("10.1234/paper", b"new content")

        path = cache.get_pdf_path("10.1234/paper")
        assert path is not None
        assert path.read_bytes() == b"new content"


class TestTextCache:
    """Tests for text caching functionality."""

    def test_has_text_false_when_not_cached(self, tmp_path):
        """Test has_text returns False for uncached text."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache.has_text("10.1234/paper") is False

    def test_has_text_true_after_save(self, tmp_path):
        """Test has_text returns True after saving text."""
        cache = PaperCache(cache_dir=tmp_path)
        cache.save_text("10.1234/paper", "Extracted text content")
        assert cache.has_text("10.1234/paper") is True

    def test_get_text_returns_none_when_not_cached(self, tmp_path):
        """Test get_text returns None for uncached text."""
        cache = PaperCache(cache_dir=tmp_path)
        assert cache.get_text("10.1234/paper") is None

    def test_get_text_returns_content_after_save(self, tmp_path):
        """Test get_text returns correct content after saving."""
        cache = PaperCache(cache_dir=tmp_path)
        text_content = "This is the extracted text from the paper."

        cache.save_text("10.1234/paper", text_content)
        retrieved = cache.get_text("10.1234/paper")

        assert retrieved == text_content

    def test_save_and_retrieve_unicode_text(self, tmp_path):
        """Test saving and retrieving Unicode text."""
        cache = PaperCache(cache_dir=tmp_path)
        unicode_text = "Unicode: \u00e9\u00e8\u00ea \u4e2d\u6587 \U0001f4da"

        cache.save_text("10.1234/paper", unicode_text)
        retrieved = cache.get_text("10.1234/paper")

        assert retrieved == unicode_text

    def test_save_text_returns_correct_path(self, tmp_path):
        """Test that save_text returns the correct path."""
        cache = PaperCache(cache_dir=tmp_path)
        path = cache.save_text("10.1234/paper", "test")

        expected_path = cache.text_dir / "10.1234_paper.txt"
        assert path == expected_path

    def test_save_text_overwrites_existing(self, tmp_path):
        """Test that save_text overwrites existing cached text."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_text("10.1234/paper", "original text")
        cache.save_text("10.1234/paper", "new text")

        retrieved = cache.get_text("10.1234/paper")
        assert retrieved == "new text"


class TestCacheClear:
    """Tests for cache clearing functionality."""

    def test_clear_removes_all_pdfs(self, tmp_path):
        """Test that clear removes all cached PDFs."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_pdf("10.1234/paper1", b"pdf1")
        cache.save_pdf("10.1234/paper2", b"pdf2")
        cache.save_pdf("10.1234/paper3", b"pdf3")

        cache.clear()

        assert cache.has_pdf("10.1234/paper1") is False
        assert cache.has_pdf("10.1234/paper2") is False
        assert cache.has_pdf("10.1234/paper3") is False
        assert list(cache.pdf_dir.glob("*.pdf")) == []

    def test_clear_removes_all_text(self, tmp_path):
        """Test that clear removes all cached text."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_text("10.1234/paper1", "text1")
        cache.save_text("10.1234/paper2", "text2")
        cache.save_text("10.1234/paper3", "text3")

        cache.clear()

        assert cache.has_text("10.1234/paper1") is False
        assert cache.has_text("10.1234/paper2") is False
        assert cache.has_text("10.1234/paper3") is False
        assert list(cache.text_dir.glob("*.txt")) == []

    def test_clear_removes_both_pdfs_and_text(self, tmp_path):
        """Test that clear removes both PDFs and text."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_pdf("10.1234/paper", b"pdf content")
        cache.save_text("10.1234/paper", "text content")

        cache.clear()

        assert cache.has_pdf("10.1234/paper") is False
        assert cache.has_text("10.1234/paper") is False

    def test_clear_preserves_directories(self, tmp_path):
        """Test that clear preserves the cache directories."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_pdf("10.1234/paper", b"pdf")
        cache.save_text("10.1234/paper", "text")

        cache.clear()

        assert cache.pdf_dir.exists()
        assert cache.text_dir.exists()

    def test_clear_on_empty_cache(self, tmp_path):
        """Test that clear works on empty cache."""
        cache = PaperCache(cache_dir=tmp_path)
        # Should not raise any errors
        cache.clear()
        assert cache.pdf_dir.exists()
        assert cache.text_dir.exists()


class TestCacheIsolation:
    """Tests for cache isolation between different paper IDs."""

    def test_different_papers_have_different_cache_entries(self, tmp_path):
        """Test that different papers have separate cache entries."""
        cache = PaperCache(cache_dir=tmp_path)

        cache.save_pdf("10.1234/paper1", b"pdf1")
        cache.save_pdf("10.1234/paper2", b"pdf2")
        cache.save_text("10.1234/paper1", "text1")
        cache.save_text("10.1234/paper2", "text2")

        # Each paper has its own cached files
        path1 = cache.get_pdf_path("10.1234/paper1")
        path2 = cache.get_pdf_path("10.1234/paper2")
        assert path1 is not None and path1.read_bytes() == b"pdf1"
        assert path2 is not None and path2.read_bytes() == b"pdf2"
        assert cache.get_text("10.1234/paper1") == "text1"
        assert cache.get_text("10.1234/paper2") == "text2"

    def test_pdf_and_text_caches_are_independent(self, tmp_path):
        """Test that PDF and text caches are independent."""
        cache = PaperCache(cache_dir=tmp_path)

        # Save only PDF
        cache.save_pdf("10.1234/paper", b"pdf content")
        assert cache.has_pdf("10.1234/paper") is True
        assert cache.has_text("10.1234/paper") is False

        # Save only text for different paper
        cache.save_text("10.1234/other", "text content")
        assert cache.has_pdf("10.1234/other") is False
        assert cache.has_text("10.1234/other") is True


class TestCacheIntegrationWithDownload:
    """Tests for cache integration with download module."""

    @pytest.fixture
    def mock_downloader(self):
        """Create a mock downloader for testing."""

        class MockDownloader(Downloader):
            name = "mock"

            def __init__(self, pdf_bytes: bytes | None = None):
                super().__init__()
                self.pdf_bytes = pdf_bytes
                self.download_calls = []

            async def download(self, doi: str) -> bytes | None:
                self.download_calls.append(doi)
                return self.pdf_bytes

            async def __aenter__(self):
                return self

            async def __aexit__(self, *args):
                pass

        return MockDownloader

    @pytest.mark.asyncio
    async def test_download_uses_cache_on_hit(self, mock_downloader, tmp_path):
        """Test that download returns from cache without calling downloader."""
        cache_dir = tmp_path / "cache"
        output_dir = tmp_path / "output"
        cache = PaperCache(cache_dir=cache_dir)

        # Pre-populate cache
        cached_content = b"%PDF-1.4 cached content"
        cache.save_pdf("10.1234/cached", cached_content)

        downloader = mock_downloader(pdf_bytes=b"new content")

        results = []
        async for result in download_papers(
            ["10.1234/cached"], output_dir, downloaders=[downloader], cache=cache
        ):
            results.append(result)

        assert len(results) == 1
        assert results[0].success is True
        assert results[0].source == "cache"
        # Downloader should not have been called
        assert len(downloader.download_calls) == 0
        # Output file should have cached content
        assert (output_dir / "10.1234_cached.pdf").read_bytes() == cached_content

    @pytest.mark.asyncio
    async def test_download_saves_to_cache_on_miss(self, mock_downloader, tmp_path):
        """Test that successful download saves PDF to cache."""
        cache_dir = tmp_path / "cache"
        output_dir = tmp_path / "output"
        cache = PaperCache(cache_dir=cache_dir)

        pdf_content = b"%PDF-1.4 downloaded content"
        downloader = mock_downloader(pdf_bytes=pdf_content)

        # Not in cache initially
        assert cache.has_pdf("10.1234/new") is False

        results = []
        async for result in download_papers(
            ["10.1234/new"], output_dir, downloaders=[downloader], cache=cache
        ):
            results.append(result)

        assert len(results) == 1
        assert results[0].success is True
        assert results[0].source == "mock"
        # Now should be in cache
        assert cache.has_pdf("10.1234/new") is True
        cached_path = cache.get_pdf_path("10.1234/new")
        assert cached_path is not None and cached_path.read_bytes() == pdf_content

    @pytest.mark.asyncio
    async def test_download_does_not_cache_on_failure(self, mock_downloader, tmp_path):
        """Test that failed download does not add to cache."""
        cache_dir = tmp_path / "cache"
        output_dir = tmp_path / "output"
        cache = PaperCache(cache_dir=cache_dir)

        downloader = mock_downloader(pdf_bytes=None)  # Returns None = failure

        results = []
        async for result in download_papers(
            ["10.1234/failed"], output_dir, downloaders=[downloader], cache=cache
        ):
            results.append(result)

        assert len(results) == 1
        assert results[0].success is False
        # Should not be in cache
        assert cache.has_pdf("10.1234/failed") is False

    @pytest.mark.asyncio
    async def test_download_with_use_cache_false(self, mock_downloader, tmp_path):
        """Test that use_cache=False bypasses cache entirely."""
        cache_dir = tmp_path / "cache"
        output_dir = tmp_path / "output"
        cache = PaperCache(cache_dir=cache_dir)

        # Pre-populate cache
        cache.save_pdf("10.1234/paper", b"cached content")

        pdf_content = b"%PDF-1.4 fresh download"
        downloader = mock_downloader(pdf_bytes=pdf_content)

        results = []
        async for result in download_papers(
            ["10.1234/paper"], output_dir, downloaders=[downloader], use_cache=False
        ):
            results.append(result)

        assert len(results) == 1
        assert results[0].success is True
        assert results[0].source == "mock"  # Not "cache"
        # Downloader was called
        assert len(downloader.download_calls) == 1
        # Output file should have fresh content
        assert (output_dir / "10.1234_paper.pdf").read_bytes() == pdf_content



import logging
import os
from collections.abc import AsyncIterator
from datetime import date
from typing import Literal
from urllib.parse import parse_qs, urlencode, urlparse

import streamish as st

from scimesh.models import Author, Paper
from scimesh.providers.base import Provider
from scimesh.query.combinators import (
    And,
    CitationRange,
    Field,
    Not,
    Or,
    Query,
    YearRange,
    extract_citation_range,
    remove_citation_range,
)

logger = logging.getLogger(__name__)


class Scopus(Provider):
    """Scopus paper search provider."""

    name = "scopus"
    BASE_URL = "https://api.elsevier.com/content/search/scopus"
    PAGE_SIZE = 25

    def _load_from_env(self) -> str | None:
        return os.getenv("SCOPUS_API_KEY")

    def _translate_query(self, query: Query) -> str:
        """Convert Query AST to Scopus query syntax."""
        match query:
            case Field(field="title", value=v):
                return f"TITLE({v})"
            case Field(field="author", value=v):
                return f"AUTH({v})"
            case Field(field="abstract", value=v):
                return f"ABS({v})"
            case Field(field="keyword", value=v):
                return f"KEY({v})"
            case Field(field="fulltext", value=v):
                return f"ALL({v})"
            case Field(field="doi", value=v):
                return f"DOI({v})"
            case And(left=l, right=r):
                return f"({self._translate_query(l)} AND {self._translate_query(r)})"
            case Or(left=l, right=r):
                return f"({self._translate_query(l)} OR {self._translate_query(r)})"
            case Not(operand=o):
                return f"NOT {self._translate_query(o)}"
            case YearRange(start=s, end=e):
                if s and e:
                    if s == e:
                        return f"PUBYEAR = {s}"
                    else:
                        return f"(PUBYEAR > {s - 1} AND PUBYEAR < {e + 1})"
                elif s:
                    return f"PUBYEAR > {s - 1}"
                elif e:
                    return f"PUBYEAR < {e + 1}"
                return ""
            case CitationRange():
                return ""
            case _:
                raise ValueError(f"Unsupported query node: {query}")

    async def search(
        self,
        query: Query,
    ) -> AsyncIterator[Paper]:
        """Search Scopus and yield papers."""
        if self._client is None:
            raise RuntimeError("Provider not initialized. Use 'async with provider:'")

        if not self._api_key:
            raise ValueError("Scopus requires an API key. Set SCOPUS_API_KEY or pass api_key=")

        client = self._client

        citation_filter = extract_citation_range(query)
        query_without_citations = remove_citation_range(query)

        if query_without_citations is None:
            return

        query_str = self._translate_query(query_without_citations)
        logger.debug("Translated query: %s", query_str)

        headers = {
            "X-ELS-APIKey": self._api_key,
            "Accept": "application/json",
        }

        sort_param: str | None = None
        if citation_filter and citation_filter.min is not None:
            sort_param = "-citedby-count"
            logger.debug("Using sort=%s for citation filter optimization", sort_param)

        async def fetch_pages() -> AsyncIterator[list[dict]]:
            cursor: str | None = "*"

            while cursor is not None:
                params: dict[str, str | int] = {
                    "query": query_str,
                    "count": self.PAGE_SIZE,
                    "view": "COMPLETE",
                    "cursor": cursor,
                }
                if sort_param:
                    params["sort"] = sort_param

                url = f"{self.BASE_URL}?{urlencode(params)}"
                logger.debug("Requesting: %s", url)
                response = await client.get(url, headers=headers)
                response.raise_for_status()

                data = response.json()
                search_results = data.get("search-results", {})
                results = search_results.get("entry", [])

                yield results

                cursor = self._extract_next_cursor(search_results)
                if len(results) < self.PAGE_SIZE:
                    break

        def passes_citation_filter(paper: Paper) -> bool:
            cit_count = paper.citations_count
            if citation_filter:
                if citation_filter.min is not None and (
                    cit_count is None or cit_count < citation_filter.min
                ):
                    return False
                if citation_filter.max is not None and (
                    cit_count is None or cit_count > citation_filter.max
                ):
                    return False
            return True

        stream = (
            st.stream(fetch_pages())
            .flat_map(lambda entries: entries)
            .map(self._parse_entry)
            .filter(lambda p: p is not None and passes_citation_filter(p))
        )
        async for paper in stream:
            if paper is not None:
                yield paper

    def _extract_next_cursor(self, search_results: dict) -> str | None:
        """Extract the next cursor from Scopus response links."""
        links = search_results.get("link", [])
        for link in links:
            if link.get("@ref") == "next":
                href = link.get("@href", "")
                if href:
                    parsed = urlparse(href)
                    query_params = parse_qs(parsed.query)
                    cursor_values = query_params.get("cursor", [])
                    if cursor_values:
                        return cursor_values[0]
        return None

    def _parse_entry(self, entry: dict) -> Paper | None:
        """Parse a Scopus entry into a Paper."""
        title = entry.get("dc:title")
        if not title:
            return None

        authors = []
        for author_data in entry.get("author", []):
            name = author_data.get("authname")
            if name:
                authors.append(Author(name=name))

        year = 0
        cover_date = entry.get("prism:coverDate", "")
        if cover_date:
            try:
                year = int(cover_date.split("-")[0])
            except (ValueError, IndexError):
                pass

        doi = entry.get("prism:doi")

        url = None
        for link in entry.get("link", []):
            if link.get("@ref") == "scopus":
                url = link.get("@href")
                break

        journal = entry.get("prism:publicationName")

        citations = None
        cited_by = entry.get("citedby-count")
        if cited_by:
            try:
                citations = int(cited_by)
            except ValueError:
                pass

        pub_date = None
        if cover_date:
            try:
                pub_date = date.fromisoformat(cover_date)
            except ValueError:
                pass

        topics = []
        subject_area = entry.get("subject-area", [])
        if isinstance(subject_area, list):
            for area in subject_area[:5]:
                if isinstance(area, dict):
                    topics.append(area.get("$", ""))

        is_oa = entry.get("openaccessFlag", False)

        pdf_url = None
        for link in entry.get("link", []):
            if link.get("@ref") == "full-text":
                pdf_url = link.get("@href")
                break

        return Paper(
            title=title,
            authors=tuple(authors),
            year=year,
            source="scopus",
            abstract=entry.get("dc:description"),
            doi=doi,
            url=url,
            topics=tuple(topics),
            citations_count=citations,
            publication_date=pub_date,
            journal=journal,
            pdf_url=pdf_url,
            open_access=is_oa,
            extras={"scopus_id": entry.get("dc:identifier")},
        )

    async def get(self, paper_id: str) -> Paper | None:
        """Fetch a specific paper by DOI or Scopus ID.

        Args:
            paper_id: DOI (e.g., "10.1038/nature14539") or Scopus ID.

        Returns:
            Paper if found, None otherwise.
        """
        if self._client is None:
            raise RuntimeError("Provider not initialized. Use 'async with provider:'")

        if not self._api_key:
            raise ValueError("Scopus requires an API key. Set SCOPUS_API_KEY or pass api_key=")

        if paper_id.startswith("SCOPUS_ID:"):
            scopus_id = paper_id.replace("SCOPUS_ID:", "")
            query_str = f"EID(2-s2.0-{scopus_id})"
        elif "/" in paper_id:
            query_str = f"DOI({paper_id})"
        else:
            query_str = f"EID(2-s2.0-{paper_id})"

        headers = {
            "X-ELS-APIKey": self._api_key,
            "Accept": "application/json",
        }

        params = {
            "query": query_str,
            "count": 1,
            "view": "COMPLETE",
        }

        url = f"{self.BASE_URL}?{urlencode(params)}"
        logger.debug("Fetching: %s", url)

        response = await self._client.get(url, headers=headers)
        response.raise_for_status()

        data = response.json()
        results = data.get("search-results", {}).get("entry", [])

        if not results:
            return None

        if "error" in results[0]:
            return None

        return self._parse_entry(results[0])

    async def citations(
        self,
        paper_id: str,
        direction: Literal["in", "out", "both"] = "both",
        max_results: int = 100,
    ) -> AsyncIterator[Paper]:
        """Get papers citing this paper.

        Note: Scopus API only supports getting papers that cite this paper (direction="in").
        For references (direction="out"), you need the Scopus Abstract API which requires
        different entitlements.

        Args:
            paper_id: DOI or Scopus ID.
            direction: Only "in" is supported (papers citing this one).
            max_results: Maximum number of results to return.

        Yields:
            Paper instances.
        """
        if self._client is None:
            raise RuntimeError("Provider not initialized. Use 'async with provider:'")

        if not self._api_key:
            raise ValueError("Scopus requires an API key. Set SCOPUS_API_KEY or pass api_key=")

        if direction == "out":
            logger.warning("Scopus does not support fetching references (direction='out')")
            return

        paper = await self.get(paper_id)
        if paper is None:
            return

        scopus_id = paper.extras.get("scopus_id")
        if not scopus_id:
            return

        if scopus_id.startswith("SCOPUS_ID:"):
            scopus_id = scopus_id.replace("SCOPUS_ID:", "")

        headers = {
            "X-ELS-APIKey": self._api_key,
            "Accept": "application/json",
        }

        params = {
            "query": f"REFEID(2-s2.0-{scopus_id})",
            "count": min(max_results, 25),
            "view": "COMPLETE",
        }

        url = f"{self.BASE_URL}?{urlencode(params)}"
        logger.debug("Fetching citing papers: %s", url)

        response = await self._client.get(url, headers=headers)
        response.raise_for_status()

        data = response.json()
        results = data.get("search-results", {}).get("entry", [])

        for entry in results:
            if "error" in entry:
                continue
            parsed = self._parse_entry(entry)
            if parsed:
                yield parsed



# tests/test_arxiv.py
import time
from unittest.mock import MagicMock

import pytest

from scimesh.providers.arxiv import ATOM_NS, Arxiv
from scimesh.query.combinators import abstract, author, fulltext, title, year


def test_translate_title():
    provider = Arxiv()
    q = title("transformer")
    result = provider._translate_query(q)
    assert result == 'ti:"transformer"'


def test_translate_author():
    provider = Arxiv()
    q = author("Vaswani")
    result = provider._translate_query(q)
    assert result == 'au:"Vaswani"'


def test_translate_abstract():
    provider = Arxiv()
    q = abstract("attention mechanism")
    result = provider._translate_query(q)
    assert result == 'abs:"attention mechanism"'


def test_translate_fulltext():
    provider = Arxiv()
    q = fulltext("neural network")
    result = provider._translate_query(q)
    assert result == 'all:"neural network"'


def test_translate_and():
    provider = Arxiv()
    q = title("BERT") & author("Google")
    result = provider._translate_query(q)
    assert result == '(ti:"BERT" AND au:"Google")'


def test_translate_or():
    provider = Arxiv()
    q = title("BERT") | title("GPT")
    result = provider._translate_query(q)
    assert result == '(ti:"BERT" OR ti:"GPT")'


def test_translate_not():
    provider = Arxiv()
    q = title("neural") & ~author("Smith")
    result = provider._translate_query(q)
    assert result == '(ti:"neural" ANDNOT au:"Smith")'


def test_translate_year_ignored():
    provider = Arxiv()
    q = year(2020, 2024)
    result = provider._translate_query(q)
    assert result == ""  # arXiv doesn't support year in query


def test_no_api_key_needed():
    provider = Arxiv()
    assert provider._api_key is None


def _make_arxiv_entry(arxiv_id: str, title_text: str) -> str:
    """Create an arXiv entry XML element."""
    return f"""
    <entry xmlns="{ATOM_NS[1:-1]}" xmlns:arxiv="http://arxiv.org/schemas/atom">
        <id>http://arxiv.org/abs/{arxiv_id}</id>
        <title>{title_text}</title>
        <published>2023-01-15T00:00:00Z</published>
        <author><name>Test Author</name></author>
        <summary>Test abstract for {title_text}</summary>
        <link href="http://arxiv.org/abs/{arxiv_id}" type="text/html"/>
        <arxiv:primary_category term="cs.AI"/>
    </entry>
    """


def _make_arxiv_response(entries: list[str], total_results: int, start: int) -> str:
    """Create a mock arXiv API response with OpenSearch metadata."""
    entries_xml = "\n".join(entries)
    return f"""<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom"
      xmlns:opensearch="http://a9.com/-/spec/opensearch/1.1/"
      xmlns:arxiv="http://arxiv.org/schemas/atom">
    <opensearch:totalResults>{total_results}</opensearch:totalResults>
    <opensearch:startIndex>{start}</opensearch:startIndex>
    <opensearch:itemsPerPage>{len(entries)}</opensearch:itemsPerPage>
    {entries_xml}
</feed>"""


@pytest.mark.asyncio
async def test_search_paginates_with_delay():
    """arXiv search should paginate with 3 second delay between requests."""
    provider = Arxiv()

    # Create 2 pages of results (100 + 50 = 150 total)
    page1_entries = [_make_arxiv_entry(f"2023.{i:05d}", f"Paper {i}") for i in range(100)]
    page2_entries = [_make_arxiv_entry(f"2023.{i:05d}", f"Paper {i}") for i in range(100, 150)]

    page1_response = _make_arxiv_response(page1_entries, 150, 0)
    page2_response = _make_arxiv_response(page2_entries, 150, 100)

    call_count = 0
    call_times: list[float] = []

    async def mock_get(url: str):
        nonlocal call_count
        call_count += 1
        call_times.append(time.monotonic())
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()

        if call_count == 1:
            mock_response.text = page1_response
        else:
            mock_response.text = page2_response

        return mock_response

    mock_client = MagicMock()
    mock_client.get = mock_get

    provider._client = mock_client

    papers = []
    async for paper in provider.search(title("test")):
        papers.append(paper)

    # Should have fetched all 150 papers across 2 pages
    assert len(papers) == 150

    # Should have made exactly 2 API calls
    assert call_count == 2

    # Should have at least 3 second delay between calls
    assert len(call_times) == 2
    delay = call_times[1] - call_times[0]
    assert delay >= 3.0, f"Expected >= 3.0 second delay, got {delay:.2f}s"


@pytest.mark.asyncio
async def test_search_single_page_no_delay():
    """arXiv search should not delay when results fit in single page."""
    provider = Arxiv()

    # Create a single page of results
    entries = [_make_arxiv_entry(f"2023.{i:05d}", f"Paper {i}") for i in range(50)]
    response = _make_arxiv_response(entries, 50, 0)

    call_count = 0

    async def mock_get(url: str):
        nonlocal call_count
        call_count += 1
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()
        mock_response.text = response
        return mock_response

    mock_client = MagicMock()
    mock_client.get = mock_get

    provider._client = mock_client

    papers = []
    async for paper in provider.search(title("test")):
        papers.append(paper)

    assert len(papers) == 50
    assert call_count == 1


@pytest.mark.asyncio
async def test_search_respects_arxiv_hard_limit(monkeypatch):
    """arXiv search should stop at 30000 (arXiv hard limit)."""
    import scimesh.providers.arxiv as arxiv_module

    # Mock asyncio.sleep to avoid waiting 15+ minutes
    async def mock_sleep(seconds):
        pass

    monkeypatch.setattr(arxiv_module.asyncio, "sleep", mock_sleep)

    provider = Arxiv()

    # Simulate a query with 35000 total results
    entries = [_make_arxiv_entry(f"2023.{i:05d}", f"Paper {i}") for i in range(100)]

    call_count = 0

    async def mock_get(url: str):
        nonlocal call_count
        call_count += 1
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()
        # Always return 35000 as total, but we should stop at 30000
        start = (call_count - 1) * 100
        mock_response.text = _make_arxiv_response(entries, 35000, start)
        return mock_response

    mock_client = MagicMock()
    mock_client.get = mock_get

    provider._client = mock_client

    papers = []
    async for paper in provider.search(title("test")):
        papers.append(paper)

    # Should stop at 30000 (300 pages * 100 per page)
    # 30000 / 100 = 300 pages
    assert call_count == 300
    assert len(papers) == 30000



import logging
import warnings
from collections.abc import AsyncIterator
from typing import Literal

import streamish as st

from scimesh.models import Paper, SearchResult
from scimesh.providers.base import Provider
from scimesh.query.combinators import Query
from scimesh.query.parser import parse

logger = logging.getLogger(__name__)

OnError = Literal["fail", "ignore", "warn"]

DEFAULT_DEDUPE_WINDOW = 10_000


async def _search_stream(
    query: Query,
    providers: list[Provider],
    on_error: OnError,
    dedupe: bool = True,
    dedupe_window: int = DEFAULT_DEDUPE_WINDOW,
) -> AsyncIterator[Paper]:
    """Stream papers from multiple providers with optional windowed deduplication."""
    logger.info("Starting search with %s providers", len(providers))

    async def safe_search(provider: Provider) -> AsyncIterator[Paper]:
        """Wrap provider search with error handling."""
        try:
            async with provider:
                async for paper in provider.search(query):
                    yield paper
        except Exception as e:
            if on_error == "fail":
                raise
            elif on_error == "warn":
                warnings.warn(f"Provider {provider.name} failed: {e}", stacklevel=3)

    def dedupe_key(paper: Paper) -> str:
        return paper.doi or f"{paper.title.lower()}:{paper.year}"

    streams = [safe_search(p) for p in providers]
    merged = st.merge(*streams)

    if dedupe:
        merged = st.distinct_by(dedupe_key, merged, window=dedupe_window)

    async for paper in merged:
        yield paper


async def collect_search(stream: AsyncIterator[Paper]) -> SearchResult:
    """Collect streamed papers into a SearchResult."""
    papers: list[Paper] = []
    totals: dict[str, int] = {}

    async for paper in stream:
        papers.append(paper)
        totals[paper.source] = totals.get(paper.source, 0) + 1

    return SearchResult(papers=papers, total_by_provider=totals)


def search(
    query: Query | str,
    providers: list[Provider],
    take: int = -1,
    on_error: OnError = "warn",
    dedupe: bool = True,
) -> AsyncIterator[Paper]:
    """
    Search for papers across multiple providers.

    Args:
        query: Query AST or Scopus-style query string
        providers: List of providers to search
        on_error: Error handling mode - "fail", "ignore", or "warn"
        dedupe: Whether to deduplicate results (uses sliding window of 10k)

    Returns:
        AsyncIterator yielding Paper objects as they arrive.

    Examples:

        async for paper in search(query, providers):
            print(paper.title)

        result = await collect_search(search(query, providers))
    """
    match query:
        case str():
            logger.debug("Parsing query string: %s", query)
            query = parse(query)

    if take < 0:
        return _search_stream(query, providers, on_error, dedupe)

    return st.take(take, _search_stream(query, providers, on_error, dedupe))



"""Playwright-based downloader for protected open access PDFs."""

from __future__ import annotations

import base64
import logging
import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Self
from urllib.parse import urlparse

from scimesh.download.base import Downloader

if TYPE_CHECKING:
    from playwright.async_api import Browser, BrowserContext, Page

    from scimesh.download.host_concurrency import HostSemaphores

logger = logging.getLogger(__name__)

try:
    from playwright.async_api import async_playwright

    PLAYWRIGHT_AVAILABLE = True
except ImportError:
    PLAYWRIGHT_AVAILABLE = False
    async_playwright = None  # type: ignore[assignment]


def _env_bool(key: str, default: bool) -> bool:
    value = os.getenv(key)
    if value is None:
        return default
    return value.lower() in ("true", "1", "yes")


def _env_int(key: str, default: int) -> int:
    value = os.getenv(key)
    if value is None:
        return default
    return int(value)


def _env_tuple(key: str) -> tuple[str, ...]:
    value = os.getenv(key)
    if not value:
        return ()
    return tuple(arg.strip() for arg in value.split(",") if arg.strip())


@dataclass(frozen=True)
class PlaywrightConfig:
    """Configuration for PlaywrightDownloader.

    All settings can be overridden via environment variables with
    the prefix SCIMESH_PLAYWRIGHT_.
    """

    browser: str = field(default_factory=lambda: os.getenv("SCIMESH_PLAYWRIGHT_BROWSER", "chrome"))
    headless: bool = field(default_factory=lambda: _env_bool("SCIMESH_PLAYWRIGHT_HEADLESS", True))
    timeout: int = field(default_factory=lambda: _env_int("SCIMESH_PLAYWRIGHT_TIMEOUT", 30))
    proxy: str | None = field(default_factory=lambda: os.getenv("SCIMESH_PLAYWRIGHT_PROXY"))
    user_agent: str | None = field(
        default_factory=lambda: os.getenv("SCIMESH_PLAYWRIGHT_USER_AGENT")
    )
    locale: str | None = field(default_factory=lambda: os.getenv("SCIMESH_PLAYWRIGHT_LOCALE"))
    viewport_width: int = field(
        default_factory=lambda: _env_int("SCIMESH_PLAYWRIGHT_VIEWPORT_WIDTH", 1920)
    )
    viewport_height: int = field(
        default_factory=lambda: _env_int("SCIMESH_PLAYWRIGHT_VIEWPORT_HEIGHT", 1080)
    )
    browser_args: tuple[str, ...] = field(
        default_factory=lambda: _env_tuple("SCIMESH_PLAYWRIGHT_BROWSER_ARGS")
    )
    stealth: bool = field(default_factory=lambda: _env_bool("SCIMESH_PLAYWRIGHT_STEALTH", True))

    @property
    def effective_args(self) -> tuple[str, ...]:
        """Browser args with stealth defaults applied."""
        args = list(self.browser_args)
        if self.stealth:
            stealth_arg = "--disable-blink-features=AutomationControlled"
            if stealth_arg not in args:
                args.append(stealth_arg)
        return tuple(args)

    @property
    def effective_user_agent(self) -> str:
        """User agent with stealth default if not specified."""
        if self.user_agent:
            return self.user_agent
        if self.stealth:
            return (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/120.0.0.0 Safari/537.36"
            )
        return ""

    @property
    def is_system_browser(self) -> bool:
        """Whether using a system-installed browser (no playwright install needed)."""
        return self.browser in ("chrome", "msedge", "chrome-beta", "msedge-beta")


class PlaywrightDownloader(Downloader):
    """Downloader that uses headless browser to bypass CDN protections.

    Some open access publishers (like MDPI) use CDN protections (Akamai, Cloudflare)
    that block direct HTTP requests. This downloader navigates to the article page
    first (to get cookies/session), then uses fetch() inside the browser context
    to download the PDF.

    Configuration can be done via PlaywrightConfig or environment variables.
    """

    name = "playwright"

    def __init__(
        self,
        config: PlaywrightConfig | None = None,
        host_semaphores: HostSemaphores | None = None,
    ):
        super().__init__(host_semaphores=host_semaphores)
        self.email = os.environ.get("UNPAYWALL_EMAIL")
        self.config = config or PlaywrightConfig()
        self._playwright: Any = None
        self._browser: Browser | None = None
        self._available = PLAYWRIGHT_AVAILABLE and bool(self.email)

        if not PLAYWRIGHT_AVAILABLE:
            logger.debug("  [Playwright] Not available: playwright not installed")
        elif not self.email:
            logger.debug("  [Playwright] Not available: UNPAYWALL_EMAIL not set")

    async def __aenter__(self) -> Self:
        await super().__aenter__()
        if self._available:
            self._playwright = await async_playwright().start()  # type: ignore[misc]
            self._browser = await self._launch_browser()
        return self

    async def _launch_browser(self) -> Browser:
        """Launch browser based on config."""
        cfg = self.config
        launch_kwargs: dict[str, Any] = {
            "headless": cfg.headless,
            "args": list(cfg.effective_args),
        }

        if cfg.proxy:
            launch_kwargs["proxy"] = {"server": cfg.proxy}

        if cfg.is_system_browser:
            launch_kwargs["channel"] = cfg.browser
            return await self._playwright.chromium.launch(**launch_kwargs)

        browser_type = cfg.browser
        if browser_type in ("chromium", "chrome"):
            return await self._playwright.chromium.launch(**launch_kwargs)
        elif browser_type == "firefox":
            return await self._playwright.firefox.launch(**launch_kwargs)
        elif browser_type == "webkit":
            return await self._playwright.webkit.launch(**launch_kwargs)
        else:
            raise ValueError(f"Unknown browser: {browser_type}")

    async def _create_context(self) -> BrowserContext:
        """Create browser context with config applied."""
        cfg = self.config
        context_kwargs: dict[str, Any] = {
            "viewport": {"width": cfg.viewport_width, "height": cfg.viewport_height},
        }

        if cfg.effective_user_agent:
            context_kwargs["user_agent"] = cfg.effective_user_agent

        if cfg.locale:
            context_kwargs["locale"] = cfg.locale

        context = await self._browser.new_context(**context_kwargs)  # type: ignore[union-attr]

        if cfg.stealth:
            await context.add_init_script(
                "Object.defineProperty(navigator, 'webdriver', { get: () => undefined });"
            )

        return context

    async def __aexit__(self, *_: object) -> None:
        if self._browser:
            await self._browser.close()
            self._browser = None
        if self._playwright:
            await self._playwright.stop()
            self._playwright = None
        await super().__aexit__(*_)

    def _unpaywall_url(self, doi: str) -> str:
        return f"https://api.unpaywall.org/v2/{doi}?email={self.email}"

    def _get_landing_page(self, pdf_url: str) -> str:
        """Extract landing page URL from PDF URL."""
        parsed = urlparse(pdf_url)
        path = parsed.path
        if path.endswith("/pdf"):
            path = path[:-4]
        elif "/pdf?" in pdf_url:
            path = path.replace("/pdf", "")
        return f"{parsed.scheme}://{parsed.netloc}{path}"

    async def _fetch_pdf_in_browser(self, page: Page, pdf_url: str) -> bytes | None:
        """Use browser's fetch() to download PDF with full context."""
        try:
            b64_content = await page.evaluate(f'''async () => {{
                try {{
                    const response = await fetch("{pdf_url}");
                    if (!response.ok) return null;
                    const blob = await response.blob();
                    return new Promise((resolve) => {{
                        const reader = new FileReader();
                        reader.onloadend = () => resolve(reader.result.split(',')[1]);
                        reader.onerror = () => resolve(null);
                        reader.readAsDataURL(blob);
                    }});
                }} catch (e) {{
                    return null;
                }}
            }}''')

            if b64_content:
                pdf_bytes = base64.b64decode(b64_content)
                if pdf_bytes[:4] == b"%PDF":
                    return pdf_bytes
                logger.debug("  [Playwright] Downloaded content is not a PDF")
            return None
        except Exception as e:
            logger.debug("  [Playwright] fetch() failed: %s", e)
            return None

    async def download(self, doi: str) -> bytes | None:
        if not self._available:
            return None

        if self._client is None or self._browser is None:
            raise RuntimeError("Downloader must be used as async context manager")

        logger.info("  [Playwright] Trying DOI: %s", doi)

        try:
            response = await self._client.get(self._unpaywall_url(doi))
            if response.status_code != 200:
                logger.debug("  [Playwright] Unpaywall returned %s", response.status_code)
                return None

            data = response.json()
            oa_locations = data.get("oa_locations", [])

            for location in oa_locations:
                pdf_url = location.get("url_for_pdf")
                if not pdf_url:
                    continue

                landing_url = self._get_landing_page(pdf_url)
                logger.debug("  [Playwright] Landing: %s", landing_url)
                logger.debug("  [Playwright] PDF URL: %s", pdf_url)

                context = await self._create_context()
                page = await context.new_page()
                try:
                    timeout_ms = self.config.timeout * 1000
                    await page.goto(landing_url, wait_until="networkidle", timeout=timeout_ms)
                    pdf_bytes = await self._fetch_pdf_in_browser(page, pdf_url)
                    if pdf_bytes:
                        logger.info("  [Playwright] Success for %s", doi)
                        return pdf_bytes
                except Exception as e:
                    logger.debug("  [Playwright] Page navigation failed: %s", e)
                finally:
                    await page.close()
                    await context.close()

        except Exception as e:
            logger.debug("  [Playwright] Error: %s", e)

        logger.info("  [Playwright] All locations failed for: %s", doi)
        return None



"""Tests for YAML workspace repository."""

from datetime import datetime

import pytest

from scimesh.workspace.models import (
    CollectionWorkspace,
    Constraints,
    ExplorationWorkspace,
    Framework,
    LogEntry,
    PaperEntry,
    PaperIndex,
    Screening,
    SearchResults,
    SLRWorkspace,
    Stats,
)
from scimesh.workspace.repository import (
    WorkspaceNotFoundError,
    YamlWorkspaceRepository,
)


class TestYamlWorkspaceRepositoryPaths:
    """Tests for repository path properties."""

    def test_index_path(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        assert repo._index_path == tmp_path / "index.yaml"

    def test_log_path(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        assert repo._log_path == tmp_path / "log.yaml"

    def test_papers_path(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        assert repo._papers_path == tmp_path / "papers.yaml"


class TestYamlWorkspaceRepositoryExists:
    """Tests for exists method."""

    def test_exists_returns_false_when_no_index(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        assert repo.exists() is False

    def test_exists_returns_true_when_index_exists(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        (tmp_path / "index.yaml").write_text("type: collection\nquestion: test\n")
        assert repo.exists() is True


class TestYamlWorkspaceRepositorySaveLoadWorkspace:
    """Tests for save and load workspace methods."""

    def test_save_and_load_slr_workspace(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        workspace = SLRWorkspace(
            question="Does exercise help?",
            framework=Framework(type="pico", fields={"population": "adults"}),
            constraints=Constraints(databases=["openalex"], year_range="2020-2024"),
            stats=Stats(total=100, included=50),
            inclusion=["RCT"],
            exclusion=["case studies"],
        )
        repo.save(workspace)
        loaded = repo.load()
        assert isinstance(loaded, SLRWorkspace)
        assert loaded.type == "slr"
        assert loaded.question == "Does exercise help?"
        assert loaded.framework.type == "pico"
        assert loaded.framework.fields["population"] == "adults"
        assert loaded.constraints.databases == ["openalex"]
        assert loaded.constraints.year_range == "2020-2024"
        assert loaded.stats.total == 100
        assert loaded.inclusion == ["RCT"]
        assert loaded.exclusion == ["case studies"]

    def test_save_and_load_exploration_workspace(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        workspace = ExplorationWorkspace(
            question="AI trends?",
            constraints=Constraints(year_range="2023-2024"),
            limit=500,
            started_at=now,
        )
        repo.save(workspace)
        loaded = repo.load()
        assert isinstance(loaded, ExplorationWorkspace)
        assert loaded.type == "exploration"
        assert loaded.question == "AI trends?"
        assert loaded.limit == 500
        assert loaded.started_at == now

    def test_save_and_load_collection_workspace(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        workspace = CollectionWorkspace(
            question="My reading list",
            stats=Stats(total=25, with_pdf=20),
        )
        repo.save(workspace)
        loaded = repo.load()
        assert isinstance(loaded, CollectionWorkspace)
        assert loaded.type == "collection"
        assert loaded.question == "My reading list"
        assert loaded.stats.total == 25
        assert loaded.stats.with_pdf == 20

    def test_load_raises_when_not_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        with pytest.raises(WorkspaceNotFoundError):
            repo.load()

    def test_save_creates_directory(self, tmp_path):
        nested = tmp_path / "nested" / "path"
        repo = YamlWorkspaceRepository(nested)
        workspace = CollectionWorkspace(question="Test")
        repo.save(workspace)
        assert nested.exists()
        assert (nested / "index.yaml").exists()


class TestYamlWorkspaceRepositoryLog:
    """Tests for log methods."""

    def test_save_and_load_log(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        entries = [
            LogEntry(
                id="search-001",
                type="search",
                query="machine learning",
                providers=["openalex"],
                executed_at=now,
                results=SearchResults(total=100, unique=80),
            ),
            LogEntry(
                id="snowball-001",
                type="snowball",
                providers=["semantic_scholar"],
                executed_at=now,
                seed_doi="10.1234/test",
                direction="backward",
            ),
        ]
        repo.save_log(entries)
        loaded = repo.load_log()
        assert len(loaded) == 2
        assert loaded[0].id == "search-001"
        assert loaded[0].type == "search"
        assert loaded[0].query == "machine learning"
        assert loaded[0].results.total == 100
        assert loaded[1].id == "snowball-001"
        assert loaded[1].seed_doi == "10.1234/test"

    def test_load_log_returns_empty_when_not_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        loaded = repo.load_log()
        assert loaded == []

    def test_append_log_creates_file(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        entry = LogEntry(
            id="search-001",
            type="search",
            executed_at=now,
        )
        repo.append_log(entry)
        loaded = repo.load_log()
        assert len(loaded) == 1
        assert loaded[0].id == "search-001"

    def test_append_log_adds_to_existing(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        entry1 = LogEntry(id="search-001", type="search", executed_at=now)
        entry2 = LogEntry(id="search-002", type="search", executed_at=now)
        repo.append_log(entry1)
        repo.append_log(entry2)
        loaded = repo.load_log()
        assert len(loaded) == 2
        assert loaded[0].id == "search-001"
        assert loaded[1].id == "search-002"


class TestYamlWorkspaceRepositoryPapers:
    """Tests for papers index methods."""

    def test_save_and_load_papers(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        papers = [
            PaperEntry(
                path="papers/2024/smith-ml",
                doi="10.1234/test1",
                title="Machine Learning Survey",
                status="included",
                search_ids=["search-001"],
            ),
            PaperEntry(
                path="papers/2023/jones-ai",
                doi="10.1234/test2",
                title="AI Review",
                status="excluded",
                search_ids=["search-002"],
            ),
        ]
        repo.save_papers(papers)
        loaded = repo.load_papers()
        assert len(loaded) == 2
        assert loaded[0].path == "papers/2024/smith-ml"
        assert loaded[0].doi == "10.1234/test1"
        assert loaded[0].status == "included"
        assert loaded[1].path == "papers/2023/jones-ai"
        assert loaded[1].title == "AI Review"

    def test_load_papers_returns_empty_when_not_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        loaded = repo.load_papers()
        assert loaded == []


class TestYamlWorkspaceRepositoryPaper:
    """Tests for individual paper methods."""

    def test_save_and_load_paper(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        paper = PaperIndex(
            title="Attention Is All You Need",
            authors=["Vaswani", "Shazeer"],
            year=2017,
            doi="10.48550/arXiv.1706.03762",
            sources=["openalex", "semantic_scholar"],
            urls={"doi": "https://doi.org/10.48550/arXiv.1706.03762"},
            tags=["transformers", "nlp"],
            citations=50000,
            journal="NeurIPS",
            open_access=True,
            pdf="attention.pdf",
            abstract="The dominant sequence transduction models...",
            screening=Screening(status="included", reason="Seminal paper", screened_at=now),
        )
        path = "papers/2017/vaswani-attention"
        repo.save_paper(path, paper)
        loaded = repo.load_paper(path)
        assert loaded.title == "Attention Is All You Need"
        assert loaded.authors == ["Vaswani", "Shazeer"]
        assert loaded.year == 2017
        assert loaded.doi == "10.48550/arXiv.1706.03762"
        assert loaded.tags == ["transformers", "nlp"]
        assert loaded.citations == 50000
        assert loaded.screening.status == "included"

    def test_save_paper_creates_directory_structure(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        paper = PaperIndex(
            title="Test Paper",
            authors=["Author"],
            year=2024,
        )
        path = "papers/2024/author-test"
        repo.save_paper(path, paper)
        expected_file = tmp_path / "papers" / "2024" / "author-test" / "index.yaml"
        assert expected_file.exists()

    def test_load_paper_raises_when_not_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        with pytest.raises(FileNotFoundError):
            repo.load_paper("papers/2024/nonexistent")

    def test_paper_exists_returns_false_when_not_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        assert repo.paper_exists("papers/2024/nonexistent") is False

    def test_paper_exists_returns_true_when_found(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        paper = PaperIndex(
            title="Test Paper",
            authors=["Author"],
            year=2024,
        )
        path = "papers/2024/author-test"
        repo.save_paper(path, paper)
        assert repo.paper_exists(path) is True


class TestYamlWorkspaceRepositoryRoundtrip:
    """Tests for full roundtrip serialization."""

    def test_full_workspace_roundtrip(self, tmp_path):
        repo = YamlWorkspaceRepository(tmp_path)
        now = datetime.now()
        workspace = SLRWorkspace(
            question="Does X affect Y?",
            framework=Framework(type="pico", fields={"population": "adults"}),
            constraints=Constraints(databases=["openalex"], year_range="2020-2024"),
            stats=Stats(total=2, included=1, excluded=1),
            inclusion=["RCT"],
            exclusion=["case study"],
        )
        repo.save(workspace)
        log_entries = [
            LogEntry(
                id="search-001",
                type="search",
                query="X AND Y",
                providers=["openalex"],
                executed_at=now,
                results=SearchResults(total=100, unique=50),
            ),
        ]
        repo.save_log(log_entries)
        papers = [
            PaperEntry(
                path="papers/2024/smith-xy",
                doi="10.1234/xy1",
                title="X affects Y",
                status="included",
                search_ids=["search-001"],
            ),
            PaperEntry(
                path="papers/2023/jones-xy",
                doi="10.1234/xy2",
                title="X does not affect Y",
                status="excluded",
                search_ids=["search-001"],
            ),
        ]
        repo.save_papers(papers)
        paper1 = PaperIndex(
            title="X affects Y",
            authors=["Smith"],
            year=2024,
            doi="10.1234/xy1",
            screening=Screening(status="included", reason="RCT", screened_at=now),
        )
        paper2 = PaperIndex(
            title="X does not affect Y",
            authors=["Jones"],
            year=2023,
            doi="10.1234/xy2",
            screening=Screening(status="excluded", reason="Case study", screened_at=now),
        )
        repo.save_paper("papers/2024/smith-xy", paper1)
        repo.save_paper("papers/2023/jones-xy", paper2)
        loaded_workspace = repo.load()
        assert isinstance(loaded_workspace, SLRWorkspace)
        assert loaded_workspace.question == "Does X affect Y?"
        loaded_log = repo.load_log()
        assert len(loaded_log) == 1
        assert loaded_log[0].query == "X AND Y"
        loaded_papers = repo.load_papers()
        assert len(loaded_papers) == 2
        loaded_paper1 = repo.load_paper("papers/2024/smith-xy")
        assert loaded_paper1.title == "X affects Y"
        assert loaded_paper1.screening.status == "included"
        loaded_paper2 = repo.load_paper("papers/2023/jones-xy")
        assert loaded_paper2.title == "X does not affect Y"
        assert loaded_paper2.screening.status == "excluded"



from __future__ import annotations

import asyncio
import logging
import re
import xml.etree.ElementTree as ET
from collections.abc import AsyncIterator
from dataclasses import replace
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urlencode

import streamish as st

from scimesh.models import Author, Paper
from scimesh.providers.base import Provider
from scimesh.query.combinators import (
    And,
    CitationRange,
    Field,
    Not,
    Or,
    Query,
    YearRange,
    extract_citation_range,
    remove_citation_range,
)

if TYPE_CHECKING:
    from scimesh.providers.openalex import OpenAlex

logger = logging.getLogger(__name__)

ATOM_NS = "{http://www.w3.org/2005/Atom}"
ARXIV_NS = "{http://arxiv.org/schemas/atom}"
OPENSEARCH_NS = "{http://a9.com/-/spec/opensearch/1.1/}"


class Arxiv(Provider):
    """arXiv paper search provider."""

    name = "arxiv"
    BASE_URL = "https://export.arxiv.org/api/query"
    PAGE_SIZE = 100
    RATE_LIMIT_DELAY = 3.0
    MAX_RESULTS = 30000

    def _load_from_env(self) -> str | None:
        return None

    def _translate_query(self, query: Query) -> str:
        """Convert Query AST to arXiv search syntax."""
        match query:
            case Field(field="title", value=v):
                return f'ti:"{v}"'
            case Field(field="author", value=v):
                return f'au:"{v}"'
            case Field(field="abstract", value=v):
                return f'abs:"{v}"'
            case Field(field="keyword", value=v):
                return f'all:"{v}"'
            case Field(field="fulltext", value=v):
                return f'all:"{v}"'
            case Field(field="doi", value=v):
                return f'doi:"{v}"'
            case And(left=l, right=r):
                left_q = self._translate_query(l)
                right_q = self._translate_query(r)
                if isinstance(r, Not):
                    return f"({left_q} ANDNOT {self._translate_query(r.operand)})"
                if not left_q:
                    return right_q
                if not right_q:
                    return left_q
                return f"({left_q} AND {right_q})"
            case Or(left=l, right=r):
                left_q = self._translate_query(l)
                right_q = self._translate_query(r)
                return f"({left_q} OR {right_q})"
            case Not(operand=o):
                return f"ANDNOT {self._translate_query(o)}"
            case YearRange():
                return ""
            case CitationRange():
                return ""
            case _:
                raise ValueError(f"Unsupported query node: {query}")

    ENRICHMENT_BATCH_SIZE = 50

    async def search(
        self,
        query: Query,
    ) -> AsyncIterator[Paper]:
        """Search arXiv and yield papers."""
        if self._client is None:
            raise RuntimeError("Provider not initialized. Use 'async with provider:'")

        citation_filter = extract_citation_range(query)
        query_without_citations = remove_citation_range(query)

        if query_without_citations is None:
            logger.warning(
                "arXiv does not provide citation data; citation-only queries return no results"
            )
            return

        if citation_filter is None:
            async for paper in self._search_raw(query_without_citations):
                yield paper
        else:
            logger.info("Enriching arXiv results with citation data from OpenAlex")
            from scimesh.providers.openalex import OpenAlex

            async with OpenAlex() as openalex:
                async for paper in (
                    st.stream(self._search_raw(query_without_citations))
                    .batch(self.ENRICHMENT_BATCH_SIZE, timeout=5.0)
                    .map_async(lambda batch: self._enrich_batch(openalex, batch), concurrency=1)
                    .flat_map(lambda papers: papers)
                    .filter(lambda p: self._passes_citation_filter(p, citation_filter))
                ):
                    yield paper

    async def _search_raw(self, query: Query) -> AsyncIterator[Paper]:
        """Raw arXiv search without citation enrichment."""
        if self._client is None:
            raise RuntimeError("Provider not initialized")

        client = self._client

        query_str = self._translate_query(query)
        logger.debug("Translated query: %s", query_str)
        if not query_str:
            logger.debug("Empty query, returning no results")
            return

        year_filter = self._extract_year_filter(query)

        async def fetch_pages() -> AsyncIterator[list[ET.Element]]:
            start = 0
            is_first_request = True

            while start < self.MAX_RESULTS:
                if not is_first_request:
                    await asyncio.sleep(self.RATE_LIMIT_DELAY)
                is_first_request = False

                params = {
                    "search_query": query_str,
                    "start": start,
                    "max_results": self.PAGE_SIZE,
                    "sortBy": "relevance",
                    "sortOrder": "descending",
                }

                url = f"{self.BASE_URL}?{urlencode(params)}"
                logger.debug("Requesting: %s", url)
                response = await client.get(url)
                response.raise_for_status()

                root = ET.fromstring(response.text)
                total_results_el = root.find(f"{OPENSEARCH_NS}totalResults")
                total_results = (
                    int(total_results_el.text)
                    if total_results_el is not None and total_results_el.text
                    else 0
                )

                entries = root.findall(f"{ATOM_NS}entry")
                yield entries

                start += self.PAGE_SIZE
                if start >= total_results or len(entries) < self.PAGE_SIZE:
                    break

        stream = (
            st.stream(fetch_pages())
            .flat_map(lambda entries: entries)
            .map(self._parse_entry)
            .filter(lambda p: p is not None and self._matches_year_filter(p, year_filter))
        )
        async for paper in stream:
            if paper is not None:
                yield paper

    async def _enrich_batch(self, openalex: OpenAlex, papers: list[Paper]) -> list[Paper]:
        """Enrich a batch of papers with citation data."""
        citations = await self._fetch_citations_batch(openalex, papers)
        return [self._with_citations(p, citations) for p in papers]

    async def _fetch_citations_batch(
        self,
        openalex: OpenAlex,
        papers: list[Paper],
    ) -> dict[str, int]:
        """Fetch citation counts from OpenAlex for arXiv papers."""
        arxiv_ids = []
        for p in papers:
            aid = p.extras.get("arxiv_id")
            if aid:
                arxiv_ids.append(aid.split("v")[0] if "v" in aid else aid)

        if not arxiv_ids:
            return {}

        dois = "|".join(f"10.48550/arXiv.{aid}" for aid in arxiv_ids)
        url = f"{openalex.BASE_URL}?filter=doi:{dois}&select=ids,cited_by_count&per_page=200"
        logger.debug("Fetching citations from OpenAlex: %s", url)

        if openalex._client is None:
            return {}
        response = await openalex._client.get(url)
        response.raise_for_status()
        data = response.json()

        result: dict[str, int] = {}
        for work in data.get("results", []):
            ids = work.get("ids", {})
            doi = ids.get("doi", "").replace("https://doi.org/", "")
            match = re.search(r"arxiv\.(\d+\.\d+)", doi, re.IGNORECASE)
            if match:
                arxiv_id = match.group(1)
                if (count := work.get("cited_by_count")) is not None:
                    result[arxiv_id] = count
        return result

    def _with_citations(self, paper: Paper, citations: dict[str, int]) -> Paper:
        """Return paper with citations_count from lookup."""
        arxiv_id = paper.extras.get("arxiv_id", "")
        arxiv_id = arxiv_id.split("v")[0] if "v" in arxiv_id else arxiv_id
        if arxiv_id in citations:
            return replace(paper, citations_count=citations[arxiv_id])
        return paper

    def _passes_citation_filter(self, paper: Paper, f: CitationRange) -> bool:
        """Check if paper passes citation filter."""
        if paper.citations_count is None:
            return False
        if f.min is not None and paper.citations_count < f.min:
            return False
        if f.max is not None and paper.citations_count > f.max:
            return False
        return True

    def _extract_year_filter(self, query: Query) -> YearRange | None:
        """Extract YearRange from query if present."""
        match query:
            case YearRange() as yr:
                return yr
            case And(left=l, right=r):
                return self._extract_year_filter(l) or self._extract_year_filter(r)
            case Or(left=l, right=r):
                return self._extract_year_filter(l) or self._extract_year_filter(r)
            case _:
                return None

    def _matches_year_filter(self, paper: Paper, year_filter: YearRange | None) -> bool:
        """Check if paper matches year filter."""
        if year_filter is None:
            return True
        if year_filter.start and paper.year < year_filter.start:
            return False
        if year_filter.end and paper.year > year_filter.end:
            return False
        return True

    def _parse_entry(self, entry: ET.Element) -> Paper | None:
        """Parse an arXiv entry XML element into a Paper."""
        title_el = entry.find(f"{ATOM_NS}title")
        if title_el is None or not title_el.text:
            return None

        title = " ".join(title_el.text.split())

        authors = []
        for author_el in entry.findall(f"{ATOM_NS}author"):
            name_el = author_el.find(f"{ATOM_NS}name")
            if name_el is not None and name_el.text:
                affil_el = author_el.find(f"{ARXIV_NS}affiliation")
                authors.append(
                    Author(
                        name=name_el.text,
                        affiliation=affil_el.text if affil_el is not None else None,
                    )
                )

        summary_el = entry.find(f"{ATOM_NS}summary")
        abstract = (
            " ".join(summary_el.text.split())
            if summary_el is not None and summary_el.text
            else None
        )

        published_el = entry.find(f"{ATOM_NS}published")
        pub_date = None
        year = 0
        if published_el is not None and published_el.text:
            try:
                pub_date = datetime.fromisoformat(published_el.text.replace("Z", "+00:00")).date()
                year = pub_date.year
            except ValueError:
                pass

        url = None
        for link in entry.findall(f"{ATOM_NS}link"):
            if link.get("type") == "text/html":
                url = link.get("href")
                break
        if not url:
            id_el = entry.find(f"{ATOM_NS}id")
            url = id_el.text if id_el is not None else None

        doi_el = entry.find(f"{ARXIV_NS}doi")
        doi = doi_el.text if doi_el is not None else None

        categories = []
        for cat in entry.findall(f"{ARXIV_NS}primary_category"):
            term = cat.get("term")
            if term:
                categories.append(term)
        for cat in entry.findall(f"{ATOM_NS}category"):
            term = cat.get("term")
            if term and term not in categories:
                categories.append(term)

        arxiv_id = None
        id_el = entry.find(f"{ATOM_NS}id")
        if id_el is not None and id_el.text:
            arxiv_id = id_el.text.split("/abs/")[-1]

        pdf_url = None
        if arxiv_id:
            base_id = arxiv_id.split("v")[0] if "v" in arxiv_id else arxiv_id
            pdf_url = f"https://arxiv.org/pdf/{base_id}.pdf"

        return Paper(
            title=title,
            authors=tuple(authors),
            year=year,
            source="arxiv",
            abstract=abstract,
            doi=doi,
            url=url,
            topics=tuple(categories),
            publication_date=pub_date,
            pdf_url=pdf_url,
            open_access=True,
            extras={"arxiv_id": arxiv_id} if arxiv_id else {},
        )

    async def get(self, paper_id: str) -> Paper | None:
        """Fetch a specific paper by arXiv ID.

        Args:
            paper_id: arXiv ID (e.g., "1908.06954" or "1908.06954v2")
                or arXiv DOI (e.g., "10.48550/arXiv.1908.06954").

        Returns:
            Paper if found, None otherwise.
        """
        if self._client is None:
            raise RuntimeError("Provider not initialized. Use 'async with provider:'")

        arxiv_id = paper_id
        if paper_id.startswith("10.48550/arXiv."):
            arxiv_id = paper_id.replace("10.48550/arXiv.", "")
        elif paper_id.startswith("10.48550/"):
            arxiv_id = paper_id.replace("10.48550/", "")

        url = f"{self.BASE_URL}?id_list={arxiv_id}"
        logger.debug("Fetching: %s", url)

        response = await self._client.get(url)
        response.raise_for_status()

        root = ET.fromstring(response.text)

        entries = root.findall(f"{ATOM_NS}entry")
        if not entries:
            return None

        entry = entries[0]

        title_el = entry.find(f"{ATOM_NS}title")
        if title_el is not None and title_el.text and "Error" in title_el.text:
            return None

        return self._parse_entry(entry)



from .arxiv import Arxiv
from .base import Provider
from .openalex import OpenAlex
from .scopus import Scopus
from .semantic_scholar import SemanticScholar

__all__ = ["Provider", "Arxiv", "OpenAlex", "Scopus", "SemanticScholar"]
